# cython: boundscheck=False, wraparound=False, cdivision=True, nonecheck=False
import numpy as np
cimport numpy as cnp
from numpy.math cimport NAN
cimport cython
from cython cimport floating
from libc.math cimport sqrt, fabs
from scipy.special.cython_special import betainc


cpdef _moments_welford(floating [:] x, floating [:] y):
    """
    Calculates means, variances, and covariance of the given input array using
    Welford's algorithm.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    mx, my, varx, vary, cov : floating
    """
    cdef int i, n
    cdef floating mx, my, mxold, myold, M2x, M2y, C
    cdef floating nobs
    n = len(x)

    mx = my = M2x = M2y = C = 0
    nobs = 0
    for i in range(n):
        nobs += 1
        mxold = mx
        myold = my
        mx += (x[i] - mx) / nobs
        my += (y[i] - my) / nobs
        M2x += (x[i] - mx) * (x[i] - mxold)
        M2y += (y[i] - my) * (y[i] - myold)
        C += (x[i] - mx) * (y[i] - myold)
    return mx, my, M2x/n, M2y/n, C/n




cpdef bias(floating [:] x, floating [:] y):
    """
    Difference of the mean values.

    Sign of output depends on argument order. We calculate mean(x) - mean(y).

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    bias : float
        Bias between x and y.
    """
    cdef floating b = 0
    cdef int i, n = len(x)
    for i in range(n):
        b += x[i] - y[i]
    return b / n


cpdef RSS(floating [:] x, floating [:] y):
    """
    Residual sum of squares.

    Parameters
    ----------
    x : numpy.ndarray
        Observations.
    y : numpy.ndarray
        Predictions.

    Returns
    -------
    res : float
        Residual sum of squares.
    """
    cdef floating sum = 0
    cdef int i
    cdef int n = len(x)
    for i in range(n):
        sum += (x[i] - y[i])**2
    return sum


cpdef mse_corr(floating [:] x, floating [:] y):
    r"""
    Correlation component of MSE.

    MSE can be decomposed into a term describing the deviation of x and y
    attributable to non-perfect correlation (r < 1), a term depending on the
    difference in variances between x and y, and the difference in means
    between x and y (bias).

    ..math::

        MSE &= MSE_{corr} + MSE_{var} + MSE_{bias}\\
            &= 2\sigma_x\sigma_y (1-r) + (\sigma_x - \sigma_y)^2
               + (\mu_x - \mu_y)^2

    This function calculates the term :math:`MSE_{corr} =
    2\sigma_x\sigma_y(1-r)`.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    mse_corr : float
        Correlation component of MSE.
    """
    cdef floating mx, my, varx, vary, cov
    mx, my, varx, vary, cov = _moments_welford(x, y)
    return _mse_corr_from_moments(mx, my, varx, vary, cov)


cpdef _mse_corr_from_moments(
    floating mx, floating my, floating vx, floating vy, floating cov
):
    return max(2 * sqrt(vx) * sqrt(vy) - 2 * cov, 0)



cpdef mse_var(floating [:] x, floating [:] y):
    r"""
    Variance component of MSE.

    MSE can be decomposed into a term describing the deviation of x and y
    attributable to non-perfect correlation (r < 1), a term depending on the
    difference in variances between x and y, and the difference in means
    between x and y (bias).

    ..math::

        MSE &= MSE_{corr} + MSE_{var} + MSE_{bias}\\
            &= 2\sigma_x\sigma_y (1-r) + (\sigma_x - \sigma_y)^2
               + (\mu_x - \mu_y)^2

    This function calculates the term :math:`MSE_{var} = (\sigma_x -
    \sigma_y)^2`.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    mse_var : float
        Variance component of MSE.
    """
    cdef floating mx, my, varx, vary, cov
    mx, my, varx, vary, cov = _moments_welford(x, y)
    return _mse_var_from_moments(mx, my, varx, vary, cov)


cpdef _mse_var_from_moments(
    floating mx, floating my, floating vx, floating vy, floating cov
):
    return (sqrt(vx) - sqrt(vy)) ** 2


cpdef mse_bias(floating [:] x, floating [:] y):
    r"""
    Bias component of MSE.

    MSE can be decomposed into a term describing the deviation of x and y
    attributable to non-perfect correlation (r < 1), a term depending on the
    difference in variances between x and y, and the difference in means
    between x and y (bias).

    ..math::

        MSE &= MSE_{corr} + MSE_{var} + MSE_{bias}\\
            &= 2\sigma_x\sigma_y (1-r) + (\sigma_x - \sigma_y)^2
               + (\mu_x - \mu_y)^2

    This function calculates the term :math:`MSE_{bias} = (\mu_x - \mu_y)^2`.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    mse_bias : float
        Bias component of MSE.
    """
    return bias(x, y) ** 2


cpdef _mse_bias_from_moments(
    floating mx, floating my, floating vx, floating vy, floating cov
):
    return (mx - my) ** 2


# Faster implementation of the old `mse`
# mse:
# 1.52 ms ± 138 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
# mse_decomposition:
# 48.3 µs ± 8.69 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
# -> 32 times faster!
cpdef mse_decomposition(floating [:] x, floating [:] y):
    r"""
    Mean square deviation/mean square error.

    For validation, MSE (same as MSE) is defined as

    ..math::

        MSE = \frac{1}{n}\sum\limits_{i=1}^n (x_i - y_i)^2

    MSE can be decomposed into a term describing the deviation of x and y
    attributable to non-perfect correlation (r < 1), a term depending on the
    difference in variances between x and y, and the difference in means
    between x and y (bias).

    ..math::

        MSE &= MSE_{corr} + MSE_{var} + MSE_{bias}\\
            &= 2\sigma_x\sigma_y (1-r) + (\sigma_x - \sigma_y)^2
               + (\mu_x - \mu_y)^2

    This function calculates the all components as well as the sum.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector

    Returns
    -------
    mse : float
        Mean square deviation
    mse_corr : float
        Correlation component of MSE.
    mse_bias : float
        Bias component of the MSE.
    mse_var : float
        Variance component of the MSE.
    """
    cdef floating mx, my, varx, vary, cov
    mx, my, varx, vary, cov = _moments_welford(x, y)

    # decompositions
    mse_corr =  _mse_corr_from_moments(mx, my, varx, vary, cov)
    mse_var =  _mse_var_from_moments(mx, my, varx, vary, cov)
    mse_bias =  _mse_bias_from_moments(mx, my, varx, vary, cov)
    mse = mse_corr + mse_var + mse_bias
    return mse, mse_corr, mse_bias, mse_var


cpdef _ubrmsd(floating [:] x, floating [:] y):
    r"""
    Unbiased root-mean-square deviation (uRMSD).

    This corresponds to RMSD with mean biases removed beforehand, that is

    ..math::

        ubRMSD = \sqrt{\frac{1}{n}\sum\limits_{i=1}^n
                           \left((x - \bar{x}) - (y - \bar{y}))^2}

    NOTE: If you are scaling the data beforehand to have zero mean bias, this
    is exactly the same as RMSD.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    ubrmsd : float
        Unbiased root-mean-square deviation (uRMSD).
    """
    cdef floating mx = 0, my = 0
    cdef floating sum = 0
    cdef int i, n = len(x)

    if n == 0:
        return NAN

    # calculate means
    for i in range(n):
        mx += x[i]
        my += y[i]
    mx /= n
    my /= n

    # get unbiased values
    for i in range(n):
        sum += ((x[i] - mx) - (y[i] - my))**2
    return sqrt(sum / n)


cpdef _pearsonr_from_moments(floating varx, floating vary, floating cov, int n):
    cdef floating R, p_R, t_squared, df, z

    # not defined in this case
    if varx == 0 or vary == 0:
        return np.nan, np.nan

    # due to rounding errors, values slightly above 1 are possible
    R = max(-1, min(cov / sqrt(varx * vary), 1))

    # p-value for R
    if fabs(R) == 1.0:
        p_R = 0.0
    else:
        df = n - 2
        t_squared = R * R * (df / ((1.0 - R) * (1.0 + R)))
        z = min(float(df) / (df + t_squared), 1.0)
        p_R = betainc(0.5*df, 0.5, z)
    return R, p_R


# This implementation is much faster than the old version with numba:
# old: 76.7 ms ± 1.62 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
# new: 117 µs ± 836 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
cpdef rolling_pr_rmsd(double [:] timestamps,
                      floating [:] x,
                      floating [:] y,
                      double window_size,
                      int center,
                      int min_periods):
    """
    Computation of rolling Pearson R and RMSD.

    Parameters
    ----------
    timestamps : float64
        Time stamps as julian dates.
    data : numpy.ndarray
        Time series data in 2d array.
    window_size : float64
        Window size in fraction of days.
    center : bool
        Set window at the center and include window_size in both directions.
    min_periods : int
        Minimum number of observations in window required for computation.

    Results
    -------
    pr_arr : numpy.array
        Rolling Pearson R and p-value.
    rmsd_arr : numpy.array
        Rolling RMSD
    """
    # This uses an adapted Welford's algorithm to calculate the rolling mean,
    # variance, covariance, and mean squared difference.
    cdef int i, j, n_ts, num_obs, lower, upper, lold, uold
    cdef floating mx, my, mxold, myold, M2x, M2y, C, msd,
    cdef floating rolling_nobs  # a float because we divide by it
    cdef cnp.ndarray[floating, ndim=2] pr_arr
    cdef cnp.ndarray[floating, ndim=1] rmsd_arr
    cdef floating [:,:] pr_view
    cdef floating [:] rmsd_view

    n_ts = len(timestamps)
    # allocate numpy arrays of the correct dtype for returning
    if floating is float:
        pr_arr = np.empty((n_ts, 2), dtype=np.float32)
        rmsd_arr = np.empty(n_ts, dtype=np.float32)
    elif floating is double:
        pr_arr = np.empty((n_ts, 2), dtype=np.float64)
        rmsd_arr = np.empty(n_ts, dtype=np.float64)
    else:
        raise ValueError("Unkown floating type")
    # work on memoryviews instead of on arrays directly
    pr_view = pr_arr
    rmsd_view = rmsd_arr

    mx = my = msd = M2x = M2y = C = 0
    lower = 0
    upper = -1
    rolling_nobs = 0
    for i in range(n_ts):
        lold = lower
        uold = upper


        # find interval
        if center:
            # find new start
            for j in range(lower, n_ts):
                lower = j
                if timestamps[j] >= timestamps[i] - window_size:
                    break

            # find new end:
            # we have to check separately whether the last entry is outside
            # the window, otherwise we will always get n_ts - 2 as result

            # if we're not at the end yet
            if timestamps[n_ts - 1] > timestamps[i] + window_size:
                if i == 0:
                    # in the first iteration upper was -1 in order to set uold
                    # correctly, but in the following upper must be at least 1,
                    # so we don't access an invalid index
                    upper = 1
                for j in range(upper, n_ts):
                    # j - 1 because we want the index that is still inside the
                    # window
                    upper = j - 1
                    if timestamps[j] > timestamps[i] + window_size:
                        break
            else:
                upper = n_ts - 1

        else:
            for j in range(lower, n_ts):
                lower = j
                if timestamps[j] > timestamps[i] - window_size:
                    break
            upper = i



        # Regardless of the number of observations we calculate the moments
        # here, because we need to update them at one point anyway

        # first, add the new terms with Welford's algorithm
        for j in range(uold+1, upper+1):
            mxold = mx
            myold = my
            rolling_nobs += 1
            mx += (x[j] - mx) / rolling_nobs
            my += (y[j] - my) / rolling_nobs
            msd += ((x[j] - y[j])**2 - msd) / rolling_nobs
            M2x += (x[j] - mx) * (x[j] - mxold)
            M2y += (y[j] - my) * (y[j] - myold)
            C += (x[j] - mx) * (y[j] - myold)

        # now subtract the ones that fell out the window
        # the old values here correspond to the m_n values in the formula,
        # that's why the order is different here
        for j in range(lold, lower):
            mxold = mx
            myold = my
            rolling_nobs -= 1
            mx -= (x[j] - mx) / rolling_nobs
            my -= (y[j] - my) / rolling_nobs
            msd -= ((x[j] - y[j])**2 - msd) / rolling_nobs
            M2x -= (x[j] - mxold) * (x[j] - mx)
            M2y -= (y[j] - myold) * (y[j] - my)
            C -= (x[j] - mxold) * (y[j] - my)

        # check if we have enough observations
        num_obs = upper - lower + 1
        if num_obs == 0 or num_obs < min_periods:
            pr_view[i, 0] = NAN
            pr_view[i, 1] = NAN
            rmsd_view[i] = NAN
        else:
            # to get var and cov we would need to divide by n, but since
            # we're only interested in the ratio that's not necessary
            pr_view[i, 0], pr_view[i, 1] = _pearsonr_from_moments(
                M2x, M2y, C, num_obs
            )
            rmsd_view[i] = sqrt(msd)

    return pr_arr, rmsd_arr



# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
Module containing utility functions that do not fit into other modules
'''
import numpy as np
import scipy.interpolate as sc_int
import scipy.optimize as sc_opt
import scipy.special as sc_special

import functools
import inspect
import warnings
import os
from pathlib import Path


def rootdir() -> Path:
    return Path(os.path.join(os.path.dirname(
        os.path.abspath(__file__)))).parents[1]


def deprecated(message: str = None):
    """
    Decorator for classes or functions to mark them as deprecated.
    If the decorator is applied without a specific message (`@deprecated()`),
    the default warning is shown when using the function/class. To specify
    a custom message use it like:
        @deprecated('Don't use this function anymore!').

    Parameters
    ----------
    message : str, optional (default: None)
        Custom message to show with the DeprecationWarning.
    """

    def decorator(src):
        default_msg = f"Pytesmo " \
                      f"{'class' if inspect.isclass(src) else 'method'} " \
                      f"'{src.__module__}.{src.__name__}' " \
                      f"is deprecated and will be removed soon."

        @functools.wraps(src)
        def new_func(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)

            warnings.warn(
                default_msg if message is None else message,
                category=DeprecationWarning,
                stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)
            return src(*args, **kwargs)

        return new_func

    return decorator


def interp_uniq(src):
    """
    replace non unique values by their linear interpolated value
    This method interpolates iteratively like it is done in IDL.

    Parameters
    ----------
    src: numpy.array
        array to ensure uniqueness of

    Returns
    -------
    src: numpy.array
        interpolated unique values in array of same size as src

    """
    size = len(src)
    uniq, uniq_ind, counts = np.unique(
        src, return_index=True, return_counts=True)

    while len(src[uniq_ind]) != size:
        # replace non unique percentiles by their linear interpolated value
        # This method interpolates iteratively like it is done in IDL
        # and might be replaced by a faster method of simple linear
        # interpolation
        for i in range(len(uniq_ind)):
            pos = np.where(src == src[uniq_ind[i]])[0]
            if len(pos) > 1:
                if pos[0] == 0 and pos[-1] < size - 1:
                    src[pos[-1]] = (src[pos[len(pos) - 2]] +
                                    src[pos[-1] + 1]) / 2.0
                elif pos[-1] == size - 1:
                    src[pos[0]] = (src[pos[1]] + src[pos[0] - 1]) / 2.0
                else:
                    src[pos[0]] = (src[pos[1]] + src[pos[0] - 1]) / 2.0
                    src[pos[1]] = (src[pos[0]] + src[pos[1] + 1]) / 2.0
            uniq_ind = np.unique(src, return_index=True)[1]

    return src


def element_iterable(el):
    """
    Test if a element is iterable

    Parameters
    ----------
    el: object


    Returns
    -------
    iterable: boolean
       if True then then el is iterable
       if Fales then not
    """
    try:
        el[0]
        iterable = True
    except (TypeError, IndexError):
        iterable = False

    return iterable


def ensure_iterable(el):
    """
    Ensure that an object is iterable by putting it into a list.
    Strings are handled slightly differently. They are
    technically iterable but we want to keep the whole.

    Parameters
    ----------
    el: object

    Returns
    -------
    iterable: list
        [el]
    """
    if type(el) == str:
        return [el]
    if not element_iterable(el):
        return [el]
    else:
        return el


def array_dropna(*arrs):
    """
    Drop elements from input arrays where ANY array is NaN

    Parameters
    ----------
    *arrs : np.array(s)
        One or multiple numpy arrays of the same length that contain nans

    Returns
    -------
    arrs_dropna : np.array
        Input arrays without NaNs
    """

    idx = ~np.logical_or(*[np.isnan(x) for x in arrs])
    arrs_dropna = [np.compress(idx, x) for x in arrs]

    if len(arrs_dropna) == 1:
        arrs_dropna = arrs_dropna[0]

    return tuple(arrs_dropna)



"""
Created on Apr 17, 2013

@author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
"""

from scipy import stats
import numpy as np
import pandas as pd
import scipy.interpolate as sc_int
from warnings import warn
import pytesmo.utils as utils
from pytesmo.cdf_matching import CDFMatching


def add_scaled(df, method="linreg", label_in=None, label_scale=None, **kwargs):
    """
    takes a dataframe and appends a scaled time series to it. If no labels are
    given the first column will be scaled to the second column of the DataFrame

    Parameters
    ----------
    df : pandas.DataFrame
        input dataframe
    method : string
        scaling method
    label_in: string, optional
        the column of the dataframe that should be scaled to that with
        label_scale default is the first column
    label_scale : string, optional
        the column of the dataframe the label_in column should be scaled to
        default is the second column

    Returns
    -------
    df : pandas.DataFrame
        input dataframe with new column labeled label_in+'_scaled_'+method
    """

    if label_in is None:
        label_in = df.columns.values[0]
    if label_scale is None:
        label_scale = df.columns.values[1]

    scaling_func = get_scaling_function(method)

    scaled = scaling_func(
        df[label_in].values, df[label_scale].values, **kwargs
    )

    new_label = label_in + "_scaled_" + method

    df[new_label] = pd.Series(scaled, index=df.index)

    return df


def scale(df, method="linreg", reference_index=0, **kwargs):
    """
    takes pandas.DataFrame and scales all columns to the column specified
    by reference_index with the chosen method

    Parameters
    ----------
    df : pandas.DataFrame
        containing matched time series that should be scaled
    method : string, optional
        method definition, has to be a function in globals() that takes 2
        numpy.array as input and returns one numpy.array of same length
    reference_index : int, optional
        default 0, column index of reference dataset in dataframe

    Returns
    -------
    scaled data : pandas.DataFrame
        all time series of the input DataFrame scaled to the one specified by
        reference_index
    """
    scaling_func = get_scaling_function(method)

    reference = df[df.columns.values[reference_index]]
    df = df.drop([df.columns.values[reference_index]], axis=1)
    # new_df = pd.DataFrame
    for series in df:
        df[series] = pd.Series(
            scaling_func(df[series].values, reference.values, **kwargs),
            index=df.index,
        )

    df.insert(reference_index, reference.name, reference)

    return df


def get_scaling_method_lut():
    """
    Get all defined scaling methods and their function names.

    Returns
    -------
    lut: dictionary
       key: scaling method name
       value: function
    """

    lut = {
        "linreg": linreg,
        "mean_std": mean_std,
        "min_max": min_max,
        "cdf_match": cdf_match,
        "cdf_beta_match": cdf_beta_match,
    }

    return lut


def get_scaling_function(method):
    """
    Get scaling function based on method name.

    Parameters
    ----------
    method: string
        method name as string

    Returns
    -------
    scaling_func: function
        function(src:numpy.ndarray, ref:numpy.ndarray) > scaled_src:np.ndarray

    Raises
    ------
    KeyError:
        if method is not found
    """
    lut = get_scaling_method_lut()
    try:
        return lut[method]
    except KeyError:
        raise KeyError(f"Scaling method {method} not found.")


def min_max(src, ref, **kwargs):
    """
    scales the input datasets so that they have the same minimum
    and maximum afterwards

    Parameters
    ----------
    src : numpy.array
        input dataset which will be scaled
    ref : numpy.array
        src will be scaled to this dataset

    Returns
    -------
    scaled dataset : numpy.array
        dataset src with same maximum and minimum as ref
    """
    return (src - np.min(src)) / (np.max(src) - np.min(src)) * (
        np.max(ref) - np.min(ref)
    ) + np.min(ref)


def linreg_stored_params(src, slope, intercept):
    """
    Scale the input data with passed correction values

    Parameters
    ----------
    src : numpy.array
        Candidate values, that are scaled
    slope : float
        Multiplicative correction value
    intercept : float
        Additive correction value

    Returns
    -------
    src_scaled : numpy.array
        The scaled input values
    """

    return np.abs(slope) * src + intercept


def linreg_params(src, ref):
    """
    Calculate additive and multiplicative correction parameters
    based on linear regression models.

    Parameters
    ----------
    src: numpy.array
        Candidate data (to which the corrections apply)
    ref : numpy.array
        Reference data (which candidate is scaled to)

    Returns
    -------
    slope : float
        Multiplicative correction value
    intercept : float
        Additive correction value
    """

    slope, intercept, r_value, p_value, std_err = stats.linregress(src, ref)

    return slope, intercept


def linreg(src, ref, **kwargs):
    """
    scales the input datasets using linear regression

    Parameters
    ----------
    src : numpy.array
        input dataset which will be scaled
    ref : numpy.array
        src will be scaled to this dataset

    Returns
    -------
    scaled dataset : numpy.array
        dataset scaled using linear regression
    """

    slope, intercept = linreg_params(src, ref)
    return linreg_stored_params(src, slope, intercept)


def mean_std(src, ref, **kwargs):
    """
    scales the input datasets so that they have the same mean
    and standard deviation afterwards

    Parameters
    ----------
    src : numpy.array
        input dataset which will be scaled
    ref : numpy.array
        src will be scaled to this dataset

    Returns
    -------
    scaled dataset : numpy.array
        dataset src with same mean and standard deviation as ref
    """
    return ((src - np.mean(src)) / np.std(src)) * np.std(ref) + np.mean(ref)


@utils.deprecated("Use the new implementation 'cdf_match' instead.")
def cdf_beta_match(*args, **kwargs):
    return cdf_match(*args, **kwargs)


def cdf_match(
        src, ref, nbins=100, minobs=20, linear_edge_scaling=True,
        percentiles=None, combine_invalid=True, max_val=None, min_val=None
):
    """
    Rescales by CDF matching.

    This calculates the empirical CDFs for source and reference dataset using a
    specified number of bins. In case of non-unique percentile values, a beta
    distribution is fitted to the CDF.
    For more robust estimation of the lower and upper bins, linear edge scaling
    is used (see Moesinger et al., 2020 for details).

    Parameters
    ----------
    src: numpy.array
        input dataset which will be scaled
    ref: numpy.array
        src will be scaled to this dataset
    nbins: int, optional
        Number of bins to use for estimation of the CDF
    percentiles : sequence, optional
        Percentile values to use. If this is given, `nbins` is ignored. The
        percentiles might still be changed if `minobs` is given and the number
        data per bin is lower. Default is ``None``.
    minobs : int, optional
        Minimum desired number of observations in a bin for bin resizing. If it
        is ``None`` bins will not be resized. Default is 20.
    linear_edge_scaling : bool, optional
        Whether to derive the edge parameters via linear regression (more
        robust, see Moesinger et al. (2020) for more info). Default is
        ``True``.
        Note that this way only the outliers in the reference (y) CDF are
        handled. Outliers in the input data (x) will not be removed and will
        still show up in the data.
    combine_invalid : bool, optional
        Optional feature to combine the masks of invalid data (NaN, Inf) of
        both source (X) and reference (y) data passed to `fit`. This only makes
        sense if X and y are both timeseries data corresponding to the same
        index. In this case, this makes sures that data is only used if values
        for X and y are available, so that seasonal patterns in missing values
        in one of them do not lead to distortions. (For example, if X is
        available the whole year, but y is only available during summer, the
        distribution of y should not be matched against the whole year CDF of
        X, because that could introduce systematic seasonal biases).
        Default is True.
    max_val, min_val : float, optional
        Maximum and minimum values to enforce.

    Returns
    -------
    CDF matched values: numpy.array
        dataset src with CDF as ref
    """
    matcher = CDFMatching(nbins=nbins, minobs=minobs,
                          linear_edge_scaling=linear_edge_scaling,
                          percentiles=percentiles,
                          combine_invalid=combine_invalid)
    matcher.fit(src, ref)
    scaled = matcher.predict(src)
    if max_val is not None:
        scaled[scaled > max_val] = max_val
    if min_val is not None:
        scaled[scaled < min_val] = min_val
    return scaled



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without
#     specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import numpy as np

from pytesmo.validation_framework.upscaling import Upscaling, MixinReadTs


class DataManager(MixinReadTs):

    """
    Class to handle the data management.

    Parameters
    ----------
    datasets : dict of dicts
        :Keys : string, datasets names
        :Values : dict, containing the following fields

            'class' : object
                Class containing the method `read` for reading the data.
            'columns' : list
                List of columns which will be used in the validation process.
            'args' : list, optional
                Args that are passed to the reading function.
            'kwargs' : dict, optional
                Kwargs that are passed to the reading function.
            'grids_compatible' : boolean, optional
                If set to True the grid point index is used directly when
                reading other, if False then lon, lat is used and a nearest
                neighbour search is necessary.
                default: False
            'use_lut' : boolean, optional
                If set to True the grid point index (obtained from a
                calculated lut between reference and other) is used when
                reading other, if False then lon, lat is used and a
                nearest neighbour search is necessary.
                default: False
            'max_dist' : float, optional
                Maximum allowed distance in meters for the lut calculation.
                default: None
    ref_name : string
        Name of the reference dataset. The reference dataset is used as spatial
        reference, i.e. all other dataset will be interpolated to the locations
        of the reference dataset.
    period : list, optional
        Of type [datetime start, datetime end]. If given then the two input
        datasets will be truncated to start <= dates <= end.
    read_ts_names : string or dict of strings, optional
        if another method name than 'read' should be used for reading the data
        then it can be specified here. If it is a dict then specify a
        function name for each dataset.
    upscale_parms : dict, optional. Default is None.
        dictionary with parameters for the upscaling methods. Keys:
            * 'upscaling_method': method for upscaling
            * 'temporal_stability': bool for using temporal stability
            * 'upscaling_lut': dict of shape
                {'other_name':{ref gpi: [other gpis]}}

    Methods
    -------
    use_lut(other_name)
        Returns lut between reference and other if use_lut for other dataset
        was set to True.
    get_result_names()
        Return results names based on reference and others names.
    read_reference(*args)
        Function to read and prepare the reference dataset.
    read_other(other_name, *args)
        Function to read and prepare the other datasets.
    """

    def __init__(
        self,
        datasets,
        ref_name,
        period=None,
        read_ts_names="read",
        upscale_parms=None,
    ):
        self.datasets = datasets
        self._add_default_values()
        self.reference_name = ref_name
        self.upscale_parms = upscale_parms

        self.other_name = []
        for dataset in datasets.keys():
            if dataset != ref_name:
                self.other_name.append(dataset)
                if "use_lut" not in self.datasets[dataset]:
                    self.datasets[dataset]["use_lut"] = False

        try:
            self.reference_grid = self.datasets[self.reference_name][
                "class"
            ].grid
        except AttributeError:
            self.reference_grid = None

        self.period = period

        # get reading functions
        if type(read_ts_names) is dict:
            self.read_ts_names = read_ts_names
        else:
            d = {}
            for dataset in datasets:
                d[dataset] = read_ts_names
            self.read_ts_names = d

        # match points in space
        if upscale_parms:
            # initialize class that performs upscaling operations
            others_class = {}
            for other in self.other_name:
                others_class[other] = self.datasets[other]["class"]
            self.luts = Upscaling(
                ref_class=datasets[self.reference_name]["class"],
                others_class=others_class,
                upscaling_lut=upscale_parms["upscaling_lut"],
                manager_parms=self.__dict__,
            )
        else:
            # combine ref to NNs only
            self.luts = self.get_luts()

    def _add_default_values(self):
        """
        Add defaults for args, kwargs, grids_compatible, use_lut and
        max_dist to dataset dictionary.
        """
        defaults = {
            "use_lut": False,
            "args": [],
            "kwargs": {},
            "grids_compatible": False,
            "max_dist": np.inf,
        }
        for dataset in self.datasets.keys():
            new_defaults = dict(defaults)
            new_defaults.update(self.datasets[dataset])
            self.datasets[dataset] = new_defaults

    def get_luts(self):
        """
        Returns luts between reference and others if use_lut for other datasets
        was set to True.

        Returns
        -------
        luts : dict
            Keys: other datasets names
            Values: lut between reference and other, or None
        """
        luts = {}
        for other_name in self.other_name:
            if self.datasets[other_name]["use_lut"]:
                luts[other_name] = self.reference_grid.calc_lut(
                    self.datasets[other_name]["class"].grid,
                    max_dist=self.datasets[other_name]["max_dist"],
                )
            else:
                luts[other_name] = None

        return luts

    @property
    def ds_dict(self):
        ds_dict = {}
        for dataset in self.datasets.keys():
            ds_dict[dataset] = self.datasets[dataset]["columns"]
        return ds_dict

    def get_results_names(self, n=2):

        return get_result_names(self.ds_dict, self.reference_name, n=n)

    def read_reference(self, *args):
        """
        Function to read and prepare the reference dataset.

        Calls read of the dataset.
        Takes either 1 (gpi) or 2 (lon, lat) arguments.

        Parameters
        ----------
        gpi : int
            Grid point index
        lon : float
            Longitude of point
        lat : float
            Latitude of point

        Returns
        -------
        ref_df : pandas.DataFrame or None
            Reference dataframe.
        """
        return self.read_ds(self.reference_name, *args)

    def read_other(self, name, *args):
        """
        Function to read and prepare non-reference datasets.

        Calls read of the dataset.

        Takes either 1 (gpi) or 2 (lon, lat) arguments.

        Parameters
        ----------
        name : string
            Name of the other dataset.
        gpi : int
            Grid point index
        lon : float
            Longitude of point
        lat : float
            Latitude of point

        Returns
        -------
        data_df : pandas.DataFrame or None
            Data DataFrame.
        """
        return self.read_ds(name, *args)

    def get_data(self, gpi, lon, lat):
        """
        Get all the data from this manager for a certain
        grid point, longitude, latidude combination.

        Parameters
        ----------
        gpi: int
            grid point indices
        lon: float
            grid point longitude
        lat: type
            grid point latitude

        Returns
        -------
        df_dict: dict of pandas.DataFrames
            Dictionary with dataset names as the key and
            pandas.DataFrames containing the data for the point
            as values.
            The dict will be empty if no data is available.
        """
        df_dict = {}

        ref_dataframe = self.read_reference(gpi)
        # if no reference data available continue with the next gpi
        if ref_dataframe is None:
            return df_dict

        other_dataframes = self.get_other_data(gpi, lon, lat)
        # if no other data available continue with the next gpi
        if len(other_dataframes) == 0:
            return df_dict

        df_dict = other_dataframes
        df_dict.update({self.reference_name: ref_dataframe})

        return df_dict

    def get_other_data(self, gpi, lon, lat):
        """
        Get all the data for non reference datasets
        from this manager for a certain
        grid point, longitude, latidude combination.

        Parameters
        ----------
        gpi: int
            grid point indices
        lon: float
            grid point longitude
        lat: type
            grid point latitude

        Returns
        -------
        other_dataframes: dict of pandas.DataFrames
            Dictionary with dataset names as the key and
            pandas.DataFrames containing the data for the point
            as values.
            The dict will be empty if no data is available.
        """

        other_dataframes = {}
        for other_name in self.other_name:
            grids_compatible = self.datasets[other_name]["grids_compatible"]
            if grids_compatible:
                other_dataframe = self.read_other(other_name, gpi)
            elif isinstance(self.luts, Upscaling):
                other_dataframe = self.luts.get_upscaled_ts(
                    gpi=gpi, other_name=other_name, **self.upscale_parms
                )
            elif self.luts[other_name] is not None:
                other_gpi = self.luts[other_name][gpi]
                if other_gpi == -1:
                    continue
                other_dataframe = self.read_other(other_name, other_gpi)
            else:
                # other_dataframe = self.read_other(other_name, lon, lat)
                max_dist = self.datasets[other_name].get("max_dist", np.inf)
                grid = self.datasets[other_name]["class"].grid
                other_gpi, dist = grid.find_nearest_gpi(
                    lon, lat, max_dist=max_dist
                )
                if np.isinf(dist):
                    # no other point found in range, currently this is handled
                    # by returning None
                    other_dataframe = None
                else:
                    other_dataframe = self.read_other(other_name, other_gpi)

            if other_dataframe is not None:
                other_dataframes[other_name] = other_dataframe
        return other_dataframes


def flatten(seq):
    eltl = []
    for elt in seq:
        t = type(elt)
        if t is tuple or t is list:
            for elt2 in flatten(elt):
                eltl.append(elt2)
        else:
            eltl.append(elt)
    return eltl


def get_result_combinations(ds_dict, n=2):
    """
    Get all possible combinations dataset columns

    Parameters
    ----------
    ds_dict: dict
       Dict of lists containing the dataset names as keys and a list of the
       columns to read from the dataset as values.
    n: int
        Number of datasets for combine with each other.
        If n=2 always two datasets will be combined into one result.
        If n=3 always three datasets will be combined into one results and
        so on.
        n has to be <= the number of total datasets.

    Returns
    -------
    results_names : list of tuples
        Containing all possible combinations of
        (dataset_x.column, dataset_y.column)
        for all datasets in ds_dict
    """
    combis = []
    for key in ds_dict:
        combis.extend(get_result_names(ds_dict, key, n=n))
    return sorted(list(set(combis)))


def get_result_names(ds_dict, refkey, n=2):
    """
    Return result names based on all possible combinations based on a
    reference dataset.

    Parameters
    ----------
    ds_dict: dict
       Dict of lists containing the dataset names as keys and a list of the
       columns to read from the dataset as values.
    refkey: string
       dataset name to use as a reference
    n: int
        Number of datasets for combine with each other.
        If n=2 always two datasets will be combined into one result.
        If n=3 always three datasets will be combined into one results and
        so on.
        n has to be <= the number of total datasets.

    Returns
    -------
    results_names : list of tuples
        Containing all combinations of
        (referenceDataset.column, otherDataset.column)
    """
    results_names = []

    ref_columns = []
    for column in ds_dict[refkey]:
        ref_columns.append((refkey, column))

    other_columns = []
    other_names = list(ds_dict)
    del other_names[other_names.index(refkey)]
    for other in sorted(other_names):
        for column in ds_dict[other]:
            other_columns.append((other, column))

    for comb in itertools.product(
        ref_columns, itertools.combinations(other_columns, n - 1)
    ):
        results_names.append(comb)

    # flatten to one level and remove those that do not have n unique
    # datasets
    results_names = flatten(results_names)

    # iterate in chunks of n*2 over the list
    result_combos = []
    for chunk in [
        results_names[pos: pos + n * 2]
        for pos in range(0, len(results_names), n * 2)
    ]:
        combo = []
        datasets = chunk[::2]
        columns = chunk[1::2]
        # if datasets are compared to themselves then don't include the
        # combination
        if len(set(datasets)) != n:
            continue
        for dataset, column in zip(datasets, columns):
            combo.append((dataset, column))
        result_combos.append(tuple(sorted(combo)))

    return result_combos



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the TU Wien, Department of Geodesy and
#      Geoinformation nor the names of its contributors may be used to endorse
#      or promote products derived from this software without specific prior
#      written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Metric Calculator Adapters change how metrics are calculated by calling
the `calc_metric` function of the adapted calculator instead of the unadapted
version.
"""

from pytesmo.validation_framework.metric_calculators import (
    PairwiseIntercomparisonMetrics,
    TripleCollocationMetrics
)
import warnings
import numpy as np


class MonthsMetricsAdapter(object):
    """
    Adapt MetricCalculators to calculate metrics for groups across months
    """

    _supported_metric_calculators = (
        PairwiseIntercomparisonMetrics,
        TripleCollocationMetrics,
    )

    def __init__(self, calculator, sets=None):
        """
        Add functionality to a metric calculator to calculate validation
        metrics for subsets of certain months in a time series (e.g. seasonal).

        Parameters
        ----------
        calculator : PairwiseIntercomparisonMetrics or TripleCollocationMetrics
            A metric calculator to adapt. Preferably an instance of a metric
             calculator listed in `_supported_metric_calculators`
        sets : dict, optional (default: None)
            Define groups of data. With group names as key and a list of
            months (1-12) that belong to the group as values.

            e.g. {'Group1': [4,5,6,7,8,9], 'Group2': [10,11,12,1,2,3]} will
            split the data used by the metric calculator into 2 groups.
            One using only observations made between April and September,
            and one using observations from the rest of the year.

            The name will be used in the results to distinguish between the
            same metrics for different groups:
            e.g. ('Group1', 'BIAS'): ..., ('Group2', 'BIAS'): ..., etc.

            The default groups are based on 4 seasons plus one group that uses
            all data (as the unadapted metric calculator would do):
            {'DJF': [12,1,2], 'MAM': [3,4,5], 'JJA': [6, 7, 8],
             'SON': [9, 10, 11], 'ALL': list(range(1, 13))}
        """
        if not isinstance(calculator, self._supported_metric_calculators):
            warnings.warn(f"Adapting {calculator.__class__} is not supported.")
        self.cls = calculator
        if sets is None:
            sets = {
                "DJF": [12, 1, 2],
                "MAM": [3, 4, 5],
                "JJA": [6, 7, 8],
                "SON": [9, 10, 11],
                "ALL": list(range(1, 13)),
            }

        self.sets = sets

        # metadata metrics and lon, lat, gpi are excluded from applying
        # seasonally
        self.non_seas_metrics = ["gpi", "lon", "lat"]
        if self.cls.metadata_template is not None:
            self.non_seas_metrics += list(self.cls.metadata_template.keys())

        all_metrics = calculator.result_template
        subset_metrics = {}

        # for each subset create a copy of the metric template
        for name in sets.keys():
            for k, v in all_metrics.items():
                if k in self.non_seas_metrics:
                    subset_metrics[k] = np.array(v)
                else:
                    subset_metrics[(name, k)] = np.array(v)

        self.result_template = subset_metrics

    @staticmethod
    def filter_months(df, months, dropna=False):
        """
        Select only entries of a time series that are within certain month(s)

        Parameters
        ----------
        df : pd.DataFrame
            Time series (index.month must exist) that is split up into the
            selected groups.
        months : list
            Months for which data is kept, e.g. [12,1,2] to keep data for
            winter
        dropna : bool, optional (default: False)
            Drop lines for months that are not to be kept, if this is false,
            the original index is not changed, but filtered values are replaced
            with nan.

        Returns
        -------
        df_filtered : pd.DataFrame
            The filtered series
        """
        dat = df.copy(True)
        dat["__index_month"] = dat.index.month
        cond = ["__index_month == {}".format(m) for m in months]
        selection = dat.query(" | ".join(cond)).index
        dat.drop("__index_month", axis=1, inplace=True)

        if dropna:
            return dat.loc[selection]
        else:
            dat.loc[dat.index.difference(selection)] = np.nan
            return dat

    def calc_metrics(self, data, gpi_info):
        """
        Calculates the desired statistics, for each set that was defined.

        Parameters
        ----------
        data : pandas.DataFrame
            with 2 columns, the first column is the reference dataset
            named 'ref'
            the second column the dataset to compare against named 'other'
        gpi_info : tuple
            Grid point info (i.e. gpi, lon, lat)
        """
        dataset = self.result_template.copy()

        for setname, months in self.sets.items():
            df = self.filter_months(data, months=months, dropna=True)
            ds = self.cls.calc_metrics(df, gpi_info=gpi_info)
            for metric, res in ds.items():
                if metric in self.non_seas_metrics:
                    k = f"{metric}"
                else:
                    k = (f"{setname}", *np.atleast_1d(metric))
                dataset[k] = res

        return dataset



# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Tue Nov  3 09:53:23 2015

Module for testing the scaling module
'''

import pytesmo.scaling as scaling
import numpy as np
import pandas as pd
import numpy.testing as nptest
import pytest


scaling_methods = list(scaling.get_scaling_method_lut().keys())


def test_mean_std_scaling():

    # use a random sample from a standard distribution
    n = 1000
    x = np.random.normal(0, 0.5, n)
    y = np.arange(n)

    o = scaling.mean_std(y, x)
    nptest.assert_almost_equal(np.std(x), np.std(o))
    nptest.assert_almost_equal(np.mean(x), np.mean(o))


def test_min_max_scaling():

    # use a random sample from a standard distribution
    n = 1000
    x = np.random.normal(0, 0.5, n)
    y = np.arange(n)

    o = scaling.min_max(y, x)
    nptest.assert_almost_equal(np.min(x), np.min(o))
    nptest.assert_almost_equal(np.max(x), np.max(o))


@pytest.mark.parametrize('method', scaling_methods)
def test_scaling_method(method):
    """
    Very simple test for linear regression.
    Could probably be done better.
    """

    # two linear functions should be matched onto each other
    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5

    o = getattr(scaling, method)(y, x)
    nptest.assert_almost_equal(x, o)


@pytest.mark.parametrize('method', scaling_methods)
def test_scaling_kwargs(method):
    """
    Tests that kwargs are passed correctly for every scaling method
    """

    # two linear functions should be matched onto each other
    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5

    kwargs = {'linear_edge_scaling': True,
              'minobs': 20,
              'max_val': 950,
              'min_val': 50}
    getattr(scaling, method)(y, x, **kwargs)


@pytest.mark.parametrize('method', scaling_methods)
def test_scale(method):

    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'])
    df_scaled = scaling.scale(df,
                              method=method,
                              reference_index=0)
    nptest.assert_almost_equal(df_scaled['x'].values,
                               df_scaled['y'].values)


@pytest.mark.parametrize('method', ['non_existing_method'])
def test_scale_error(method):

    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'])
    with pytest.raises(KeyError):
        df_scaled = scaling.scale(df,
                                  method=method,
                                  reference_index=0)
        nptest.assert_almost_equal(df_scaled['x'].values,
                                   df_scaled['y'].values)


@pytest.mark.parametrize('method', ['non_existing_method'])
def test_add_scale_error(method):

    n = 1000
    x = np.arange(n, dtype=float)
    y = np.arange(n) * 0.5

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'])
    with pytest.raises(KeyError):
        df_scaled = scaling.add_scaled(df, method=method)
        nptest.assert_almost_equal(df_scaled['y'].values,
                                   df_scaled['x_scaled_' + method].values)


@pytest.mark.parametrize('method', scaling_methods)
def test_add_scale(method):

    n = 1000
    x = np.arange(n, dtype=float)
    y = np.arange(n) * 0.5

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'])
    df_scaled = scaling.add_scaled(df, method=method)
    # test the scaling the other way round
    df_scaled2 = scaling.add_scaled(df, method=method,
                                    label_in='y',
                                    label_scale='x')
    nptest.assert_almost_equal(df_scaled['y'].values,
                               df_scaled['x_scaled_' + method].values)

    nptest.assert_almost_equal(df_scaled2['x'].values,
                               df_scaled2['y_scaled_' + method].values)


def test_linreg_with_nan():
    n = 1000
    x = np.arange(float(n))
    y = np.arange(float(n)) * 0.5

    x[0:10] = np.nan

    df = pd.DataFrame(data={'x': x, 'y': y})

    slope, inter = scaling.linreg_params(
        df.dropna()['x'].values, df.dropna()['y'].values
    )
    df['x'] = scaling.linreg_stored_params(df['x'].values, slope, inter)

    nptest.assert_almost_equal(df.loc[10:, 'x'].values,
                               df.loc[10:, 'y'].values)

    assert(df.index.size == n)


def test_single_percentile_data():

    n = 1000
    x = np.arange(n, dtype=float)
    y = np.ones(n)

    s = scaling.cdf_match(y, x)
    nptest.assert_almost_equal(s, np.full_like(s, np.nan))



# Copyright (c) 2017,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
Tests for the data scalers.
'''

from pytesmo.validation_framework.data_scalers import CDFStoreParamsScaler
from pygeogrids.grids import genreg_grid
import numpy as np
import pandas as pd


def get_dataframe_2_col():
    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5

    df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'])
    return df


def get_dataframe_3_col():
    n = 1000
    x = np.arange(n)
    y = np.arange(n) * 0.5
    z = np.arange(n) * 0.2 + 5

    df = pd.DataFrame({'x': x, 'y': y, 'z': z}, columns=['x', 'y', 'z'])
    return df


def test_CDF_storage_scaler_scale_df_2_columns(tmpdir):

    df = get_dataframe_2_col()
    path = str(tmpdir)
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    scaled_data = scaler.scale(df, 0, (1, 10, 10))
    np.testing.assert_almost_equal(scaled_data['x'].values,
                                   scaled_data['y'].values)


def test_CDF_storage_scaler_scale_df_3_columns(tmpdir):

    df = get_dataframe_3_col()
    path = str(tmpdir)
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    scaled_data = scaler.scale(df, 0, (1, 10, 10))
    np.testing.assert_almost_equal(scaled_data['x'].values,
                                   scaled_data['y'].values)
    np.testing.assert_almost_equal(scaled_data['x'].values,
                                   scaled_data['z'].values)


def test_CDF_storage_scaler_scale_df_3_columns_load_precomputed(tmpdir):

    df = get_dataframe_3_col()
    path = str(tmpdir)
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    scaler.get_parameters(df, 0, 1)
    scaler.get_parameters(df, 0, 100)
    scaled_data = scaler.scale(df, 0, (1, 10, 10))
    np.testing.assert_almost_equal(scaled_data['x'].values,
                                   scaled_data['y'].values)
    np.testing.assert_almost_equal(scaled_data['x'].values,
                                   scaled_data['z'].values)


def test_CDF_storage_scaler_calc_parameters():

    df = get_dataframe_2_col()
    path = 'test'
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    parameters = scaler.calc_parameters(df, 0)
    parameters_should = {'y_x': np.array([
        [0., 24.75, 49.75, 149.75, 249.75, 349.75, 449.75, 474.75, 499.5],
        [0., 49.5, 99.5, 299.5, 499.5, 699.5, 899.5, 949.5, 999.]
    ])}
    assert sorted(
        list(parameters_should.keys())) == sorted(list(parameters.keys()))
    for key in parameters_should:
        np.testing.assert_allclose(parameters[key],
                                   parameters_should[key], atol=1e-10)


def test_CDF_storage_scaler_store_load_parameters(tmpdir):

    path = str(tmpdir)
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    data_written = {'y_x': np.array([
        [0., 24.75, 49.75, 149.75, 249.75, 349.75, 449.75, 474.75, 499.5],
        [0., 49.5, 99.5, 299.5, 499.5, 699.5, 899.5, 949.5, 999.]
    ])}
    scaler.store_parameters(1, data_written)
    params_loaded = scaler.load_parameters(1)
    assert sorted(
        list(data_written.keys())) == sorted(list(params_loaded.keys()))
    for key in params_loaded:
        np.testing.assert_allclose(params_loaded[key],
                                   data_written[key], rtol=1e-10)


def test_CDF_storage_scaler_get_parameters(tmpdir):

    df = get_dataframe_2_col()
    path = str(tmpdir)
    grid = genreg_grid(1, 1).to_cell_grid(10, 10)
    scaler = CDFStoreParamsScaler(path, grid)
    data_written = {'y_x': np.array([
        [0., 24.75, 49.75, 149.75, 249.75, 349.75, 449.75, 474.75, 499.5],
        [0., 49.5, 99.5, 299.5, 499.5, 699.5, 899.5, 949.5, 999.]
    ])}
    params_loaded = scaler.get_parameters(df, 0, 1)
    assert sorted(
        list(data_written.keys())) == sorted(list(params_loaded.keys()))
    for key in params_loaded:
        np.testing.assert_allclose(params_loaded[key],
                                   data_written[key], atol=1e-10)

    # ensure that get_parameters has also stored them
    params_loaded = scaler.load_parameters(1)
    assert sorted(
        list(data_written.keys())) == sorted(list(params_loaded.keys()))
    for key in params_loaded:
        np.testing.assert_allclose(params_loaded[key],
                                   data_written[key], atol=1e-10)



import pytest

from src.pytesmo.validation_framework.adapters import TimestampAdapter
"""
Test for the adapters.
"""

from datetime import datetime
import numpy as np
import numpy.testing as nptest
import os
import pandas as pd
import warnings

from pytesmo.validation_framework.adapters import (MaskingAdapter,
                                                   AdvancedMaskingAdapter,
                                                   ColumnCombineAdapter)
from pytesmo.validation_framework.adapters import SelfMaskingAdapter
from pytesmo.validation_framework.adapters import AnomalyAdapter
from pytesmo.validation_framework.adapters import AnomalyClimAdapter
from tests.test_validation_framework.test_datasets import TestDataset

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from ascat.read_native.cdr import AscatGriddedNcTs


def test_masking_adapter():
    for col in (None, "x"):
        ds = TestDataset("", n=20)
        ds_mask = MaskingAdapter(ds, "<", 10, col)
        data_masked = ds_mask.read()

        nptest.assert_almost_equal(
            data_masked["x"].values,
            np.concatenate(
                [np.ones((10), dtype=bool),
                 np.zeros((10), dtype=bool)]),
        )

        if col is None:
            nptest.assert_almost_equal(data_masked["y"].values,
                                       np.ones((20), dtype=bool))


def test_self_masking_adapter():
    ref_x = np.arange(10)
    ref_y = np.arange(10) * 0.5
    ds = TestDataset("", n=20)

    ds_mask = SelfMaskingAdapter(ds, "<", 10, "x")
    data_masked = ds_mask.read()

    nptest.assert_almost_equal(data_masked["x"].values, ref_x)
    nptest.assert_almost_equal(data_masked["y"].values, ref_y)


def my_bitmasking(a, b):
    return a.astype(int) & b == b


def test_advanced_masking_adapter():
    ref_x = np.arange(5, 15, 2)
    ref_y = np.arange(5, 15, 2) * 0.5
    ds = TestDataset("", n=20)

    ds_mask = AdvancedMaskingAdapter(
        ds,
        [
            ("x", ">=", 5),
            ("x", "<", 15),
            ("x", my_bitmasking, 1),
        ],
    )
    data_masked = ds_mask.read()

    nptest.assert_almost_equal(data_masked["x"].values, ref_x)
    nptest.assert_almost_equal(data_masked["y"].values, ref_y)

    # 9 is not a valid operator, should raise an exception
    with pytest.raises(ValueError):
        ds_mask = AdvancedMaskingAdapter(
            ds,
            [
                ("x", 9, 5),
            ],
        )
        _ = ds_mask.read()


def test_advanced_masking_adapter_nans_ignored():
    ds = TestDataset("", n=20)
    # introduce nan
    ts = ds.read()
    ts.iloc[7]["x"] = np.nan

    def _read():
        return ts

    setattr(ds, "read", _read)

    # the NaN in the flag field (x) is filtered out normally
    ds_mask = AdvancedMaskingAdapter(
        ds,
        [
            ("x", ">=", 5),
            ("x", "<", 15),
        ],
    )

    data_masked = ds_mask.read()

    ref_x = np.array([5., 6., 8., 9., 10., 11., 12., 13., 14.])
    ref_y = ref_x * 0.5

    nptest.assert_almost_equal(data_masked["x"].values, ref_x)
    nptest.assert_almost_equal(data_masked["y"].values, ref_y)

    # the NaN is now ignored
    ds_mask = AdvancedMaskingAdapter(
        ds,
        [
            ("x", ">=", 5),
            ("x", "<", 15),
        ],
        ignore_nans=True,
    )

    data_masked = ds_mask.read()

    ref_x = np.array([5., 6., np.nan, 8., 9., 10., 11., 12., 13., 14.])
    ref_y = np.arange(5, 15) * 0.5

    nptest.assert_almost_equal(data_masked["x"].values, ref_x)
    nptest.assert_almost_equal(data_masked["y"].values, ref_y)


def test_anomaly_adapter():
    ds = TestDataset("", n=20)
    ds_anom = AnomalyAdapter(ds)
    data_anom = ds_anom.read()
    nptest.assert_almost_equal(data_anom["x"].values[0], -8.5)
    nptest.assert_almost_equal(data_anom["y"].values[0], -4.25)


def test_anomaly_adapter_one_column():
    ds = TestDataset("", n=20)
    ds_anom = AnomalyAdapter(ds, columns=["x"])
    data_anom = ds_anom.read()
    nptest.assert_almost_equal(data_anom["x"].values[0], -8.5)
    nptest.assert_almost_equal(data_anom["y"].values[0], 0)


def test_anomaly_clim_adapter():
    ds = TestDataset("", n=20)
    ds_anom = AnomalyClimAdapter(ds)
    data_anom = ds_anom.read()
    nptest.assert_almost_equal(data_anom["x"].values[4], -5.5)
    nptest.assert_almost_equal(data_anom["y"].values[4], -2.75)


def test_anomaly_clim_adapter_one_column():
    ds = TestDataset("", n=20)
    ds_anom = AnomalyClimAdapter(ds, columns=["x"])
    data_anom = ds_anom.read()
    nptest.assert_almost_equal(data_anom["x"].values[4], -5.5)
    nptest.assert_almost_equal(data_anom["y"].values[4], 2)


def test_adapters_custom_fct_name():

    def assert_all_read_fcts(reader):
        assert (np.all(reader.read() == reader.read()))
        assert (np.all(reader.read() == reader.alias_read()))

    base = TestDataset("", n=20)
    assert_all_read_fcts(base)
    sma = SelfMaskingAdapter(
        base, '>=', 5, column_name='y', read_name='alias_read')
    assert_all_read_fcts(sma)
    smanom = AnomalyAdapter(sma, read_name='alias_read')
    assert_all_read_fcts(smanom)


# the ascat reader gives back an ascat timeseries instead of a dataframe -
# make sure the adapters can deal with that
def test_adapters_with_ascat():
    ascat_data_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "sat",
        "ascat",
        "netcdf",
        "55R22",
    )
    ascat_grid_folder = os.path.join(
        os.path.dirname(__file__),
        "..",
        "test-data",
        "sat",
        "ascat",
        "netcdf",
        "grid",
    )
    grid_fname = os.path.join(ascat_grid_folder, "TUW_WARP5_grid_info_2_1.nc")

    ascat_reader = AscatGriddedNcTs(
        ascat_data_folder,
        "TUW_METOP_ASCAT_WARP55R22_{:04d}",
        grid_filename=grid_fname,
    )

    ascat_anom = AnomalyAdapter(ascat_reader, window_size=35, columns=["sm"])
    data = ascat_anom.read(12.891455, 45.923004)
    assert data is not None
    assert np.any(data["sm"].values != 0)

    ascat_self = SelfMaskingAdapter(ascat_reader, ">", 0, "sm")
    data2 = ascat_self.read(12.891455, 45.923004)
    assert data2 is not None
    assert np.all(data2["sm"].values > 0)
    data2 = ascat_self.read(12.891455, 45.923004)
    assert data2 is not None
    assert np.all(data2["sm"].values > 0)

    ascat_mask = MaskingAdapter(ascat_reader, ">", 0, "sm")
    data3 = ascat_mask.read(12.891455, 45.923004)
    assert data3 is not None
    assert np.any(data3["sm"].values)
    data3 = ascat_mask.read(12.891455, 45.923004)
    assert data3 is not None
    assert np.any(data3["sm"].values)

    ascat_clim = AnomalyClimAdapter(ascat_reader, columns=["sm"])
    data4 = ascat_clim.read(12.891455, 45.923004)
    assert data4 is not None
    assert np.any(data["sm"].values != 0)
    data4 = ascat_clim.read(12.891455, 45.923004)
    assert data4 is not None
    assert np.any(data["sm"].values != 0)


class TestTimezoneReader(object):

    def read(self, *args, **kwargs):
        data = np.arange(5.0)
        data[3] = np.nan
        return pd.DataFrame(
            {"data": data},
            index=pd.date_range(
                datetime(2007, 1, 1, 0), "2007-01-05", freq="D", tz="UTC"),
        )

    def read_ts(self, *args, **kwargs):
        return self.read(*args, **kwargs)


@pytest.mark.filterwarnings("ignore:Dropping timezone information:UserWarning")
def test_timezone_removal():
    tz_reader = TestTimezoneReader()

    reader_anom = AnomalyAdapter(tz_reader, window_size=35, columns=["data"])
    assert reader_anom.read(0) is not None

    reader_self = SelfMaskingAdapter(tz_reader, ">", 0, "data")
    assert reader_self.read(0) is not None

    reader_mask = MaskingAdapter(tz_reader, ">", 0, "data")
    assert reader_mask.read(0) is not None

    reader_clim = AnomalyClimAdapter(tz_reader, columns=["data"])
    assert reader_clim.read(0) is not None


def test_column_comb_adapter():
    ds = TestDataset("", n=20)
    orig = ds.read()
    ds_adapted = ColumnCombineAdapter(
        ds,
        func=pd.DataFrame.mean,
        columns=["x", "y"],
        func_kwargs={'skipna': True},
        new_name='xy_mean')
    ds_mean = ds_adapted.read()

    nptest.assert_equal(ds_mean["x"].values, orig["x"].values)
    nptest.assert_equal(ds_mean["y"].values, orig["y"].values)
    nptest.assert_equal(ds_mean["xy_mean"].values,
                        (ds_mean["x"] + ds_mean["y"]) / 2.)

    # try an empty DataFrame
    def read_empty():
        return pd.DataFrame(columns=['x', 'y'])

    setattr(ds, "read", read_empty)
    ds_adapted = ColumnCombineAdapter(
        ds,
        func=pd.DataFrame.mean,
        columns=["x", "y"],
        func_kwargs={'skipna': True},
        new_name='xy_mean')

    pd.testing.assert_frame_equal(ds_adapted.read(),
                                  pd.DataFrame(columns=['x', 'y', 'xy_mean']))


def test_timestamp_adapter():
    ds = TestDataset("", n=20)

    # Simple case
    # ================

    index = np.arange('2005-02', '2005-03', dtype='datetime64[D]')
    sm_var = np.random.randn(*index.shape)
    time_offset_field = np.random.normal(
        loc=1000.0, scale=1.0, size=index.shape).astype(int)

    def _read():
        return pd.DataFrame(
            data=np.array([sm_var, time_offset_field]).transpose(),
            columns=["sm", "offset"],
            index=index)

    setattr(ds, "read", _read)
    origin = ds.read().drop(columns="offset")

    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="s")
    adapted = adapted_ds.read()

    # Date should be unchanges as we are using a ~1000 sec offset
    assert (origin.index.date == adapted.index.date).all()
    # The offset is expressed in seconds
    assert origin.index[0] + np.timedelta64(time_offset_field[0],
                                            "s") == adapted.index[0]
    # The dataframe is integral
    assert (origin.columns == adapted.columns).all()

    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="m")
    adapted = adapted_ds.read()

    # The offset is expressed in minutes
    assert origin.index[0] + np.timedelta64(time_offset_field[0],
                                            "m") == adapted.index[0]

    # This time we do not drop the columns
    origin = ds.read()
    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="s", drop_original=False)
    adapted = adapted_ds.read()

    # The dataframe is integral
    assert (origin.columns == adapted.columns).all()

    # test use of new column
    # -----------------------
    adapted_ds = TimestampAdapter(
        ds,
        time_offset_fields="offset",
        time_units="s",
        replace_index=False,
        output_field="exact_timestamp")
    adapted = adapted_ds.read()
    assert (adapted.columns == ["sm", "exact_timestamp"]).all()

    # test NaNs in offset and generic time
    # -----------------------
    index = np.arange('2005-02', '2005-03', dtype='datetime64[D]')
    index[4] = np.datetime64("NaT")
    sm_var = np.random.randn(*index.shape)
    time_offset_field = np.random.normal(
        loc=1000.0, scale=1.0, size=index.shape)
    time_offset_field[7] = np.nan

    def _read_nans():
        return pd.DataFrame(
            data=np.array([sm_var, time_offset_field]).transpose(),
            columns=["sm", "offset"],
            index=index)

    setattr(ds, "read", _read_nans)
    origin = ds.read()

    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="s")
    adapted = adapted_ds.read()

    # One index (NaT) value should be dropped
    assert len(adapted.index) == len(origin.index) - 1
    # The Nan offset should be interpreted as 0
    assert adapted.index[6] == datetime(2005, 2, 8, 0, 0)

    # test all NaNs in dataframe
    # -----------------------
    index = np.arange('2005-02', '2005-03', dtype='datetime64[D]')
    index[:] = np.datetime64("NaT")
    sm_var = np.random.randn(*index.shape)
    time_offset_field = np.random.normal(
        loc=1000.0, scale=1.0, size=index.shape)
    time_offset_field[7] = np.nan

    def _read_all_nans():
        return pd.DataFrame(
            data=np.array([sm_var, time_offset_field]).transpose(),
            columns=["sm", "offset"],
            index=index)

    setattr(ds, "read", _read_all_nans)
    origin = ds.read()

    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="s")
    adapted = adapted_ds.read()

    # The original should be returned
    pd.testing.assert_frame_equal(origin, adapted)

    # test empty Dataframe
    # -----------------------

    def _read_empty():
        return pd.DataFrame(columns=["sm", "offset"],)

    setattr(ds, "read", _read_empty)
    origin = ds.read()

    adapted_ds = TimestampAdapter(
        ds, time_offset_fields="offset", time_units="s")
    adapted = adapted_ds.read()

    # The original should be returned
    assert adapted.empty
    assert (adapted.columns == ["sm", "offset"]).all()

    # Complex case
    # ================
    base_time = np.arange(100, 200)
    index = base_time.copy()
    sm_var = np.random.randn(*index.shape)
    time_offset_field_min = np.random.normal(
        loc=40.0, scale=1.0, size=index.shape).astype(int)
    time_offset_field_sec = np.random.normal(
        loc=1000.0, scale=1.0, size=index.shape).astype(int)

    def _read_complex():
        return pd.DataFrame(
            data=np.array([
                sm_var, base_time, time_offset_field_min, time_offset_field_sec
            ]).transpose(),
            columns=["sm", "base_time", "offset_min", "offset_sec"],
            index=index)

    setattr(ds, "read", _read_complex)
    origin = ds.read()

    adapted_ds = TimestampAdapter(
        ds,
        time_offset_fields=["offset_min", "offset_sec"],
        time_units=["m", "s"],
        base_time_field="base_time",
        base_time_reference="2005-02-01")

    adapted = adapted_ds.read()

    should_be = origin.apply(
        lambda row: np.datetime64("2005-02-01") + np.timedelta64(
            int(row["base_time"]), "D") + np.timedelta64(
                int(row["offset_min"]), "m") + np.timedelta64(
                    int(row["offset_sec"]), "s"),
        axis=1).values

    assert (adapted.index.values == should_be).all()



# Copyright (c) 2021, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the TU Wien, Department of Geodesy and
#      Geoinformation nor the names of its contributors may be used to endorse
#      or promote products derived from this software without specific prior
#      written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import warnings
from typing import Union

import pandas as pd
from scipy.stats import linregress

from pygeobase.object_base import TS
from pytesmo.temporal_matching import combined_temporal_collocation


class MixinReadTs:
    """Mixin class to provide the reading function in DataAverager
    and DataManager"""

    def read_ds(self, name, *args):
        """
        Function to read and prepare a datasets.

        Calls read_ts of the dataset.

        Takes either 1 (gpi) or 2 (lon, lat) arguments.

        Parameters
        ----------
        name : string
            Name of the other dataset.
        args : either gpi or (lon, lat)
            * gpi (int): Grid point index
            * lon (float): Longitude of point
            * lat(float): Latitude of point

        Returns
        -------
        data_df : pandas.DataFrame or None
            Data DataFrame.

        """
        ds = self.datasets[name]
        args = list(args)
        args.extend(ds["args"])

        try:
            func = getattr(ds["class"], self.read_ts_names[name])
            data_df = func(*args, **ds["kwargs"])
            if type(data_df) is TS or issubclass(type(data_df), TS):
                data_df = data_df.data
        except IOError:
            warnings.warn(
                "IOError while reading dataset {} with args {:}".format(
                    name, args
                )
            )
            return None
        except RuntimeError as e:
            if e.args[0] == "No such file or directory":
                warnings.warn(
                    "IOError while reading dataset {} with args {:}".format(
                        name, args
                    )
                )
                return None
            else:
                raise e

        if len(data_df) == 0:
            warnings.warn("No data for dataset {}".format(name))
            return None

        if not isinstance(data_df, pd.DataFrame):
            warnings.warn("Data is not a DataFrame {:}".format(args))
            return None

        if self.period is not None:
            # here we use the isoformat since pandas slice behavior is
            # different when using datetime objects.
            data_df = data_df[
                self.period[0].isoformat(): self.period[1].isoformat()
            ]

        if len(data_df) == 0:
            warnings.warn(
                "No data for dataset {} with arguments {:}".format(name, args)
            )
            return None

        else:
            return data_df


class Upscaling(MixinReadTs):
    """
    This class provides methods to combine the measurements of validation
    datasets (others) that fall under the same gridpoint of the dataset being
    validated (reference).

    The goal is to include here all identified upscaling methods to provide
    an estimate at the reference footprint scale.

    Implemented methods:

        * time-stability filtering
        * simple averaging

    Parameters
    ----------
    ref_class : <reader object> of the reference
        Class containing the method read_ts for reading the data of
        the reference
    others_class : dict
        Dict of shape {'other_name': <reader object>} for the other dataset
    upscaling_lut : dict
        Dict of shape {'other_name':{ref gpi: [other gpis]}}
    manager_parms : dict
        Dict of DataManager attributes
    """

    def __init__(
        self,
        ref_class,
        others_class,
        upscaling_lut,
        manager_parms,
    ):
        self.ref_class = ref_class
        self.others_class = others_class
        self.lut = upscaling_lut
        # attributes used by the Mixin class:
        self.datasets = manager_parms["datasets"]
        self.period = manager_parms["period"]
        self.read_ts_names = manager_parms["read_ts_names"]

    def _read(
        self,
        points,
        other_name,
    ) -> list:
        """
        Get the timeseries for given points info and return them in a list.

        Parameters
        ----------
        points : list of tuples
            list of tuples of (gpi, lon, lat)
        other_name : str
            Name of the dataset which the points belong to

        Returns
        -------
        dss : list
            list of dataframes of the reference timeseries
        """
        dss = []
        for gpi, lon, lat in points:
            # todo: check attributes in loop
            ds = self.read_ds(other_name, gpi)
            dss.append(ds)

        return dss

    @staticmethod
    def temporal_match(
        to_match, hours=6, drop_missing=False, **kwargs
    ) -> pd.DataFrame:
        """
        Temporal match to the longest timeseries

        Parameters
        ----------
        to_match : list
            list of dataframes to match
        hours : int
            window to perform the temporal matching
        drop_missing : bool, optional. Default is False.
            If true, only time steps when all points have measurements
            are kept

        Returns
        -------
        matched: pd.DataFrame
            dataframe with temporally matched timeseries
        """
        # get time series with most points
        ref = to_match[0]
        for n, df in enumerate(to_match):
            if df is None:
                continue
            points = int(df.count())
            if int(ref.count()) >= points:
                continue
            else:
                ref = df

        combined_dropna = False
        if drop_missing:
            combined_dropna = "any"

        matched = combined_temporal_collocation(
            ref,
            to_match,
            pd.Timedelta(hours, "H"),
            combined_dropna=combined_dropna,
            checkna=True,
        )
        matched.dropna(axis="columns", how="all", inplace=True)

        return matched

    @staticmethod
    def tstability_filter(
        df,
        r_min=0.6,
        see_max=0.05,
        min_n=4,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Uses time stability concepts to filter point-measurements (pms).
        Determines whether the upscaled measurement based on a simple average
        of all the pms is in sufficient agreement with each pm, and if not
        eliminates pm from the pool.

        Thresholds are based on:
        Wagner W, Pathe C, Doubkova M, Sabel D, Bartsch A, Hasenauer S,
        Blöschl G, Scipal K, Martínez-Fernández J, Löw A.
        Temporal Stability of Soil Moisture and Radar Backscatter Observed
        by the Advanced Synthetic Aperture Radar (ASAR). Sensors. 2008;
        8(2):1174-1197. https://doi.org/10.3390/s80201174

        Parameters
        ----------
        df : pd.DataFrame
            temporally matched DataFrame
        r_min : float
            lower threshold for correlation (Pearson) between upscaled and pm
        see_max : float
            upper threshold for standard error of estimate
        min_n : int
            minimum number of pms to perform the filtering

        Returns
        -------
        filtered : pd.DataFrame
            filtered input
        """
        if len(df.columns) < min_n:
            return df

        # get a trivial (average) upscale estimate
        estimate = df.mean(axis=1)

        filter_out = []
        for n, pm in enumerate(df):
            pm_values = df[pm]
            regr = linregress(estimate, pm_values)
            if regr.rvalue < r_min or regr.stderr > see_max:
                filter_out.append(pm)

        filtered = df.drop(filter_out, axis="columns")

        return filtered

    def upscale(self, df, method="average", **kwargs) -> pd.Series:
        """
        Handle the column names and return the upscaled Dataframe with the
        specified method.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe of values to upscale using method
        method : str
            averaging method
        kwargs : keyword arguments
            Arguments for some upscaling functions

        Returns
        -------
        upscaled : pandas.DataFrame
            dataframe with "upscaled" column
        """
        # New upscaling methods can be specified here in the lookup table
        up_function = {
            "average": self._average,
        }
        if method not in up_function.keys():
            raise KeyError(
                "The selected method {} is not implemented in the "
                "upscaling options".format(
                    method
                )
            )
        f = up_function[method]

        return f(df, **kwargs)

    @staticmethod
    def _average(df, **kwargs) -> pd.Series:
        """Simple average of each column in the dataframe"""
        out = df.mean(axis=1)

        return out

    def get_upscaled_ts(
        self,
        gpi,
        other_name,
        upscaling_method="average",
        temporal_stability=False,
        **kwargs,
    ) -> Union[None, pd.DataFrame]:
        """
        Find the upscale estimate timeseries with given method, for a certain
        reference gpi

        Parameters
        ----------
        gpi : int
            gpi value of the reference point
        other_name : str
            name of the non-reference dataset to be upscaled
        upscaling_method : str
            method to use for upscaling:
                * 'average' takes the simple mean of all timeseries
        temporal_stability : bool, default is False
            if True, the values are filtered using the time stability concept
        kwargs : keyword arguments
            arguments for the temporal window or time stability thresholds

        Returns
        -------
        upscaled : pd.DataFrame or None
            upscaled time series; if there are no points under the specific
            gpi, None is returned
        """
        other_lut = self.lut[other_name]
        # check that there are points for specific reference gpi
        if gpi not in other_lut.keys():
            return None
        else:
            other_points = other_lut[gpi]

        # read non-reference points and filter out Nones
        tss = self._read(points=other_points, other_name=other_name)
        tss = [df for df in tss if df is not None]

        # handle situation with single timeseries or all None
        if len(tss) <= 1:
            if not tss:
                return None
            return tss[0]

        # here we collect only the variable columns; flags are irrelevant at
        # this point and can be dropped
        target_column = self.datasets[other_name]["columns"][0]
        to_match = []
        for n, point_df in enumerate(tss):
            point_ts = point_df[target_column]
            to_match.append(
                point_ts.to_frame(
                    name=target_column + "_{}".format(n)
                )  # avoid name clashing
            )

        tss = self.temporal_match(to_match, **kwargs)
        if temporal_stability:
            tss = self.tstability_filter(tss, **kwargs)

        # perform upscaling and return correct name
        upscaled = self.upscale(tss, method=upscaling_method).to_frame(
            target_column
        )

        return upscaled



# -*- coding: utf-8 -*-
# Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, 
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
This example program loops through all insitu stations that measure soil moisture with a depth between 0 and 0.1m
it then finds the nearest ASCAT grid point and reads the ASCAT data. After temporal matching and scaling using linear
CDF matching it computes several metrics, like the correlation coefficients(Pearson's, Spearman's and Kendall's), Bias,
RMSD as well as the Nash–Sutcliffe model efficiency coefficient.


Created on Aug 8, 2013

@author: Christoph Paulik Christoph.Paulik@geo.tuwien.ac.at
'''

import ismn.interface as ismn
import ascat
import pytesmo.temporal_matching as temp_match
import pytesmo.scaling as scaling
import pytesmo.df_metrics as df_metrics
import pytesmo.metrics as metrics

import os
import matplotlib.pyplot as plt

testdata_folder = '/pytesmo/testdata'

ascat_data_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/55R22')
ascat_grid_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/grid')
static_layers_folder = os.path.join(testdata_folder,
                                    'sat/h_saf/static_layer')

# init the ASCAT SSM reader with the paths
ascat_SSM_reader = ascat.AscatSsmCdr(ascat_data_folder, ascat_grid_folder,
                                     grid_filename='TUW_WARP5_grid_info_2_1.nc',
                                     static_layer_path=static_layers_folder)
ascat_SSM_reader.read_bulk = True


# set path to ISMN data
ismn_data_folder = os.path.join(testdata_folder,
                                 'ismn/multinetwork/header_values')

# Initialize reader
ISMN_reader = ismn.ISMN_Interface(ismn_data_folder)


i = 0

label_ascat = 'sm'
label_insitu = 'insitu_sm'

# this loops through all stations that measure soil moisture
for station in ISMN_reader.stations_that_measure('soil moisture'):
    
    # this loops through all time series of this station that measure soil moisture
    # between 0 and 0.1 meters
    for ISMN_time_series in station.data_for_variable('soil moisture', min_depth=0, max_depth=0.1):

        ascat_time_series = ascat_SSM_reader.read(ISMN_time_series.longitude,
                                                  ISMN_time_series.latitude,
                                                  mask_ssf=True,
                                                  mask_frozen_prob=80,
                                                  mask_snow_prob=80)

        # focus only on the relevant variable
        ascat_time_series.data = ascat_time_series.data[label_ascat]

        # drop nan values before doing any matching
        ascat_time_series.data = ascat_time_series.data.dropna()

        ISMN_time_series.data = ISMN_time_series.data.dropna()
        
        # rename the soil moisture column in ISMN_time_series.data to insitu_sm
        # to clearly differentiate the time series when they are plotted together
        ISMN_time_series.data.rename(columns={'soil moisture':label_insitu}, inplace=True)
        
        # get ISMN data that was observerd within +- 1 hour(1/24. day) of the ASCAT observation
        # do not include those indexes where no observation was found
        matched_data = temp_match.matching(ascat_time_series.data, ISMN_time_series.data,
                                                window=1 / 24.)
        # matched ISMN data is now a dataframe with the same datetime index
        # as ascat_time_series.data and the nearest insitu observation      
        
        # continue only with relevant columns
        matched_data = matched_data[[label_ascat, label_insitu]]
        
        # the plot shows that ISMN and ASCAT are observed in different units
        matched_data.plot(figsize=(15, 5), secondary_y=[label_ascat],
                          title='temporally merged data')
        plt.show()
        
        # this takes the matched_data DataFrame and scales all columns to the 
        # column with the given reference_index, in this case in situ 
        scaled_data = scaling.scale(matched_data, method='lin_cdf_match',
                                         reference_index=1)
        
        # now the scaled ascat data and insitu_sm are in the same space    
        scaled_data.plot(figsize=(15, 5), title='scaled data')
        plt.show()
        
        plt.scatter(scaled_data[label_ascat].values, scaled_data[label_insitu].values)
        plt.xlabel(label_ascat)
        plt.ylabel(label_insitu)
        plt.show()
        
        # calculate correlation coefficients, RMSD, bias, Nash Sutcliffe
        x, y = scaled_data[label_ascat].values, scaled_data[label_insitu].values
        
        print("ISMN time series:", ISMN_time_series)
        print("compared to")
        print(ascat_time_series)
        print("Results:")
        
        # df_metrics takes a DataFrame as input and automatically
        # calculates the metric on all combinations of columns
        # returns a named tuple for easy printing
        print(df_metrics.pearsonr(scaled_data))
        print("Spearman's (rho,p_value)", metrics.spearmanr(x, y))
        print("Kendalls's (tau,p_value)", metrics.kendalltau(x, y))
        print(df_metrics.kendalltau(scaled_data))
        print(df_metrics.rmsd(scaled_data))
        print("Bias", metrics.bias(x, y))
        print("Nash Sutcliffe", metrics.nash_sutcliffe(x, y))
        
        
    i += 1

    # only show the first 2 stations, otherwise this program would run a long time
    # and produce a lot of plots
    if i >= 2:
        break





# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Tests for the time series filters
'''
import pytesmo.time_series.filters as filters
import numpy as np


def test_exp_filter():
    """
    Test exponential filter
    """
    test_jd = np.arange(10, dtype=np.double)
    test_data = np.array(
        [0.5, 2, 3, 4, -999999.0, 6, 7, 8, 9, np.nan], dtype=np.double)

    filtered = filters.exp_filter(test_data, test_jd, ctime=5)

    np.testing.assert_allclose(filtered, [0.5 , 1.324751, 1.997798, 2.656881,
                                          np.nan, 3.757916, 4.687961, 5.547327,
                                          6.378209, np.nan], rtol=1e-5)

def test_exp_filter_first_nan():
    """
    Test exponential filter with first value being a nan value
    """
    test_jd = np.arange(11, dtype=np.double)
    test_data = np.array(
        [np.nan, 0.5, 2, 3, 4, -999999.0, 6, 7, 8, 9, np.nan], dtype=np.double)

    filtered = filters.exp_filter(test_data, test_jd, ctime=5)

    np.testing.assert_allclose(filtered, [np.nan, 0.5 , 1.324751, 1.997798, 2.656881,
                                          np.nan, 3.757916, 4.687961, 5.547327,
                                          6.378209, np.nan], rtol=1e-5)

def test_boxcar_filter():
    """
    Test boxcar filter
    """
    test_jd = np.arange(10, dtype=np.double)
    test_data = np.array(
        [1, 2, 3, 4, -999999.0, 6, 7, 8, 9, np.nan], dtype=np.double)

    filtered = filters.boxcar_filter(test_data, test_jd, window=5)

    np.testing.assert_allclose(filtered, [2., 2.5, 2.5, 3.75, np.nan, 6.25,
                                          7.5, 7.5, 8., np.nan])


def test_boxcar_filter_arrsize_1():
    """
    Test boxcar filter with array of size 1
    """
    test_jd = np.arange(1, dtype=np.double)
    test_data = np.array(
        [1], dtype=np.double)

    filtered = filters.boxcar_filter(test_data, test_jd, window=5)

    np.testing.assert_allclose(filtered, [1.])



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
Tests for julian date conversion.
'''

import numpy as np
import numpy.testing as nptest
from datetime import datetime
from pytesmo.timedate.julian import julday
from pytesmo.timedate.julian import caldat
from pytesmo.timedate.julian import julian2date
from pytesmo.timedate.julian import julian2datetime
from pytesmo.timedate.julian import doy


def test_julday():
    jd = julday(5, 25, 2016, 10, 20, 11)
    jd_should = 2457533.9306828701
    nptest.assert_almost_equal(jd, jd_should)


def test_julday_arrays():
    jds = julday(np.array([5, 5]),
                 np.array([25, 25]),
                 np.array([2016, 2016]),
                 np.array([10, 10]),
                 np.array([20, 20]),
                 np.array([11, 11]))
    jds_should = np.array([2457533.93068287,
                           2457533.93068287])
    nptest.assert_almost_equal(jds, jds_should)


def test_julday_single_arrays():
    jds = julday(np.array([5]),
                 np.array([25]),
                 np.array([2016]),
                 np.array([10]),
                 np.array([20]),
                 np.array([11]))
    jds_should = np.array([2457533.93068287])
    nptest.assert_almost_equal(jds, jds_should)


def test_caldat():
    month, day, year = caldat(2457533.93068287)
    assert month == 5
    assert day == 25
    assert year == 2016


def test_caldat_array():
    month, day, year = caldat(np.array([2457533.93068287,
                                        2457533.93068287]))
    nptest.assert_almost_equal(month, np.array([5, 5]))
    nptest.assert_almost_equal(day, np.array([25, 25]))
    nptest.assert_almost_equal(year, np.array([2016, 2016]))


def test_julian2date():
    (year, month, day,
     hour, minute, second, micro) = julian2date(2457533.9306828701)
    assert type(year) == int
    assert year == 2016
    assert month == 5
    assert day == 25
    assert hour == 10
    assert minute == 20
    assert second == 10
    assert micro == 999976
    (year, month, day,
     hour, minute, second, micro) = julian2date(2454515.40972)
    assert year == 2008
    assert month == 2
    assert day == 18
    assert hour == 21
    assert minute == 49
    assert second == 59
    assert micro == 807989


def test_julian2date_single_array():
    (year, month, day,
     hour, minute, second, micro) = julian2date(np.array([2457533.9306828701]))
    assert type(year) == np.ndarray
    assert year == 2016
    assert month == 5
    assert day == 25
    assert hour == 10
    assert minute == 20
    assert second == 10
    assert micro == 999976


def test_julian2date_array():
    (year, month, day,
     hour, minute, second, micro) = julian2date(np.array([2457533.9306828701,
                                                          2457533.9306828701]))
    nptest.assert_almost_equal(year, np.array([2016, 2016]))
    nptest.assert_almost_equal(month, np.array([5, 5]))
    nptest.assert_almost_equal(day, np.array([25, 25]))
    nptest.assert_almost_equal(hour, np.array([10, 10]))
    nptest.assert_almost_equal(minute, np.array([20, 20]))
    nptest.assert_almost_equal(second, np.array([10, 10]))
    nptest.assert_almost_equal(micro, np.array([999976, 999976]))


def test_julian2datetime():
    dt = julian2datetime(2457533.9306828701)
    dt_should = datetime(2016, 5, 25, 10, 20, 10, 999976)
    assert dt == dt_should
    dt = julian2datetime(2457173.8604166666)
    dt_should = datetime(2015, 5, 31, 8, 39)
    assert dt == dt_should


def test_julian2datetime_single_array():
    dt = julian2datetime(np.array([2457533.9306828701]))
    dt_should = np.array([datetime(2016, 5, 25, 10, 20, 10, 999976)])
    assert type(dt) == np.ndarray
    assert np.all(dt == dt_should)


def test_julian2datetime_array():
    dt = julian2datetime(np.array([2457533.9306828701,
                                   2457533.9306828701]))
    dts = datetime(2016, 5, 25, 10, 20, 10, 999976)
    dt_should = np.array([dts, dts])
    assert type(dt) == np.ndarray
    assert np.all(dt == dt_should)


def test_doy():
    day_of_year = doy(1, 28)
    assert day_of_year == 28
    day_of_year = doy(2, 29)
    assert day_of_year == 31 + 29
    day_of_year = doy(3, 1, year=2004)
    assert day_of_year == 31 + 29 + 1
    # test numpy arrays as input
    days = np.array([28, 29, 1], dtype=int)
    months = np.array([1, 2, 3])
    days_of_year = doy(months, days, year=np.array([2005, 2004, 2004]))
    nptest.assert_allclose(days_of_year, np.array([28,
                                                   31 + 29,
                                                   31 + 29 + 1]))

    days_of_year = doy(months, days, year=2004)
    nptest.assert_allclose(days_of_year, np.array([28,
                                                   31 + 29,
                                                   31 + 29 + 1]))



"""
Definition of exceptions and error return codes.

The error codes are integer values. They should normally be non-negative if
properly handled. An error code of -1 indicates failure in handling the error.
0 indicates success.
"""

# This is the default value, but it should normally not appear anywhere
# outside, because the more specific failures should set a new return code
UNCAUGHT = -1
# Everything worked as planned
OK = 0
# If the metrics calculation (metrics_calculator.calc_metrics) fails because
# there is not enough data (len(data) < minobs)
INSUFFICIENT_DATA = 1
# If the metrics calculation fails due to another (unforeseen) reason. Should
# not happen too often.
METRICS_CALCULATION_FAILED = 2
# Failure in temporal matching
TEMPORAL_MATCHING_FAILED = 3
# Temporal matching returned without error, but the data we need is not
# available. This can happen if there is no temporal overlap of the required
# datasets, or if there is no data at all for one of the required datasets.
NO_TEMP_MATCHED_DATA = 4
# the scaling failed
SCALING_FAILED = 5
# the call to perform_validation failed due to other unforeseen reasons
VALIDATION_FAILED = 6
# the call to self.data_manager.get_data returned no data
NO_GPI_DATA = 7
# The call to self.data_manager.get_data failed
DATA_MANAGER_FAILED = 8


class ValidationError(Exception):
    # base exception, should not be used directly, only to catch all the
    # different subclasses at once
    return_code = UNCAUGHT


class MetricsCalculationError(ValidationError):
    return_code = METRICS_CALCULATION_FAILED


class TemporalMatchingError(ValidationError):
    return_code = TEMPORAL_MATCHING_FAILED


class NoTempMatchedDataError(ValidationError):
    return_code = NO_TEMP_MATCHED_DATA


class ScalingError(ValidationError):
    return_code = SCALING_FAILED


class ValidationFailedError(ValidationError):
    return_code = VALIDATION_FAILED


class NoGpiDataError(ValidationError):
    return_code = NO_GPI_DATA


class DataManagerError(ValidationError):
    return_code = DATA_MANAGER_FAILED



# Copyright (c) 2013,Vienna University of Technology, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Example script reads and plots ASCAT H25 SSM data with different masking options
and also converts the data to absolute values using the included
porosity data.


Created on Nov 8, 2013

@author: Christoph Paulik Christoph.Paulik@geo.tuwien.ac.at
'''

import matplotlib.pyplot as plt
import ascat.tuw as tuw
import os

# See also: https://github.com/TUW-GEO/ascat/blob/master/docs/read_tuw_ascat.rst

testdata_folder = '/pytesmo/testdata'

ascat_data_folder = os.path.join(testdata_folder,
                                 'sat/ascat/SSM')
ascat_grid_folder = os.path.join(testdata_folder,
                                 'sat/ascat/grid')
advisory_flags_folder = os.path.join(testdata_folder,
                                 'sat/ascat/advisory_flags')

# # init the ASCAT_SSM reader with the paths

ascat_SSM_reader = tuw.Ascat_SSM(ascat_data_folder, ascat_grid_folder,
                                 advisory_flags_path=advisory_flags_folder)

gpi = 2329253

ascat_ts = ascat_SSM_reader.read_ssm(gpi)

ascat_ts.plot()
plt.show()

# the ASCATTimeSeries object also contains metadata

print("Topographic complexity", ascat_ts.topo_complex)
print("Wetland fraction", ascat_ts.wetland_frac)
print("Porosity from GLDAS model", ascat_ts.porosity_gldas)
print("Porosity from Harmonized World Soil Database", ascat_ts.porosity_hwsd)

# It is also possible to automatically convert the soil moisture to absolute values using

ascat_ts_absolute = ascat_SSM_reader.read_ssm(gpi, absolute_values=True)
# this introduces 4 new columns in the returned data
# scaled sm and sm_noise with porosity_gldas
# scaled sm and sm_noise with porosity_hwsd
#print(ascat_ts_absolute.data)

# select relevant columns and plot
# ascat_ts_absolute.data = ascat_ts_absolute.data[['sm_por_gldas', 'sm_por_hwsd']]
ascat_ts_absolute.plot()
plt.show()

# We can also automatically mask the data during reading
# In this example all measurements where the Surface State Flag
# shows frozen or where the frozen or snow probabilities are more
# than 10 percent are removed from the time series
ascat_ts = ascat_SSM_reader.read_ssm(gpi, mask_ssf=True,
                                     mask_frozen_prob=10,
                                     mask_snow_pro=10)

ascat_ts.plot()
plt.show()




# coding: utf-8

# # The pytesmo validation framework
# 
# The pytesmo validation framework takes care of iterating over datasets, spatial and temporal matching as well as
# scaling. It uses metric calculators to then calculate metrics that are returned to the user. There are several
# metrics calculators included in pytesmo but new ones can be added simply.
# 
# ## Overview
# 
# How does the validation framework work? It makes these assumptions about the used datasets:
# 
# - The dataset readers that are used have a `read_ts` method that can be called either by a grid point index (gpi)
#   which can be any indicator that identifies a certain grid point or by using longitude and latitude. This means that
#   both call signatures `read_ts(gpi)` and `read_ts(lon, lat)` must be valid. Please check the
#   [pygeobase](https://github.com/TUW-GEO/pygeobase) documentation for more details on how a fully compatible dataset
#   class should look. But a simple `read_ts` method should do for the validation framework. This assumption can be
#   relaxed by using the `read_ts_names` keyword in the pytesmo.validation_framework.data_manager.DataManager class.
# - The `read_ts` method returns a pandas.DataFrame time series.
# - Ideally the datasets classes also have a `grid` attribute that is a
#   [pygeogrids](http://pygeogrids.readthedocs.org/en/latest/) grid. This makes the calculation of lookup tables easily
#   possible and the nearest neighbor search faster.
# 
# Fortunately these assumptions are true about the dataset readers included in pytesmo. 
# 
# It also makes a few assumptions about how to perform a validation. For a comparison study it is often necessary to
# choose a spatial reference grid, a temporal reference and a scaling or data space reference.
# 
# ### Spatial reference
# The spatial reference is the one to which all the other datasets are matched spatially. Often through nearest
# neighbor search. The validation framework uses grid points of the dataset specified as the spatial reference to
# spatially match all the other datasets with nearest neighbor search. Other, more sophisticated spatial matching
# algorithms are not implemented at the moment. If you need a more complex spatial matching then a preprocessing of
# the data is the only option at the moment.
# 
# ### Temporal reference
# The temporal reference is the dataset to which the other dataset are temporally matched. That means that the
# nearest observation to the reference timestamps in a certain time window is chosen for each comparison dataset.
# This is by default done by the temporal matching module included in pytesmo. How many datasets should be matched to
# the reference dataset at once can be configured, we will cover how to do this later.
# 
# ### Data space reference
# It is often necessary to bring all the datasets into a common data space by using. Scaling is often used for that
# and pytesmo offers a choice of several scaling algorithms (e.g. CDF matching, min-max scaling, mean-std scaling,
# triple collocation based scaling). The data space reference can also be chosen independently from the other two
# references.
# 
# ## Data Flow
# 
# After it is initialized, the validation framework works through the following steps:
# 
# 1. Read all the datasets for a certain job (gpi, lon, lat)
# 2. Read all the masking dataset if any
# 3. Mask the temporal reference dataset using the masking data
# 4. Temporally match all the chosen combinations of temporal reference and other datasets
# 5. Turn the temporally matched time series over to the metric calculators
# 6. Get the calculated metrics from the metric calculators
# 7. Put all the metrics into a dictionary by dataset combination and return them.
# 
# ## Masking datasets
# Masking datasets can be used if the datasets that are compared do not contain the necessary information to mask
#  them. For example we might want to use modelled soil temperature data to mask our soil moisture observations
# before comparing them. To be able to do that we just need a Dataset that returns a pandas.DataFrame with one column
#  of boolean data type. Everywhere where the masking dataset is `True` the data will be masked.
# 
# Let's look at a first example.
# 
# ## Example soil moisture validation: ASCAT - ISMN
# 
# This example shows how to setup the pytesmo validation framework to perform a comparison between ASCAT and ISMN data. 
# 

# In[1]:


import os

import pytesmo.validation_framework.metric_calculators as metrics_calculators

from datetime import datetime

from ascat.read_native.cdr import AscatSsmCdr
from ismn.interface import ISMN_Interface
from pytesmo.validation_framework.validation import Validation
from pytesmo.validation_framework.results_manager import netcdf_results_manager

# You need the test data from https://github.com/TUW-GEO/pytesmo-test-data for this example

# In[2]:


testdata_folder = '/pytesmo/testdata'
output_folder = '/pytesmo/code/examples/output'


# First we initialize the data readers that we want to use. In this case the ASCAT soil moisture time series and in
# situ data from the ISMN.
# 
# Initialize ASCAT reader

# In[3]:


ascat_data_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/55R22')
ascat_grid_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/grid')
static_layers_folder = os.path.join(testdata_folder,
                                    'sat/h_saf/static_layer')

ascat_reader = AscatSsmCdr(ascat_data_folder, ascat_grid_folder,
                           grid_filename='TUW_WARP5_grid_info_2_1.nc',
                           static_layer_path=static_layers_folder)
ascat_reader.read_bulk = True


# Initialize ISMN reader

# In[4]:


ismn_data_folder = os.path.join(testdata_folder,
                                 'ismn/multinetwork/header_values')

ismn_reader = ISMN_Interface(ismn_data_folder)


# The validation is run based on jobs. A job consists of at least three lists or numpy arrays specifing the grid
# point index, its latitude and longitude. In the case of the ISMN we can use the `dataset_ids` that identify every
# time series in the downloaded ISMN data as our grid point index. We can then get longitude and latitude from the
# metadata of the dataset.
# 
# **DO NOT CHANGE** the name ***jobs*** because it will be searched during the parallel processing!

# In[5]:


jobs = []

ids = ismn_reader.get_dataset_ids(variable='soil moisture', min_depth=0, max_depth=0.1)
for idx in ids:
    metadata = ismn_reader.metadata[idx]
    jobs.append((idx, metadata['longitude'], metadata['latitude']))

print("Jobs (gpi, lon, lat):")
print(jobs)


# For this small test dataset it is only one job
# 
# It is important here that the ISMN reader has a read_ts function that works by just using the `dataset_id`. In this
#  way the validation framework can go through the jobs and read the correct time series.

# In[6]:


data = ismn_reader.read_ts(ids[0])
print('ISMN data example:')
print(data.head())


# ## Initialize the Validation class
# 
# The Validation class is the heart of the validation framework. It contains the information about which datasets to
# read using which arguments or keywords and if they are spatially compatible. It also contains the settings about
# which metric calculators to use and how to perform the scaling into the reference data space. It is initialized in
# the following way:

# In[7]:

datasets = {
    'ISMN': {
        'class': ismn_reader,
        'columns': ['soil moisture']
    },
    'ASCAT': {
        'class': ascat_reader,
        'columns': ['sm'],
        'kwargs': {'mask_frozen_prob': 80,
                   'mask_snow_prob': 80,
                   'mask_ssf': True}
    }}

# The datasets dictionary contains all the information about the datasets to read. The `class` is the dataset class
# to use which we have already initialized. The `columns` key describes which columns of the dataset interest us for
# validation. This a mandatory field telling the framework which other columns to ignore. In this case the columns
# `soil moisture_flag` and `soil moisture_orig_flag` will be ignored by the ISMN reader. We can also specify
# additional keywords that should be given to the `read_ts` method of the dataset reader. In this case we want the
# ASCAT reader to mask the ASCAT soil moisture using the included frozen and snow probabilities as well as the SSF.
# There are also other keys that can be used here. Please see the documentation for explanations.

# In[8]:


period = [datetime(2007, 1, 1), datetime(2014, 12, 31)]
basic_metrics = metrics_calculators.BasicMetrics(other_name='k1')

process = Validation(
    datasets, 'ISMN',
    temporal_ref='ASCAT',
    scaling='lin_cdf_match',
    scaling_ref='ASCAT',   
    metrics_calculators={(2, 2): basic_metrics.calc_metrics},
    period=period)


# During the initialization of the Validation class we can also tell it other things that it needs to know. In this
# case it uses the datasets we have specified earlier. The spatial reference is the `'ISMN'` dataset which is the
# second argument. The 'metrics_calculators' argument looks a little bit strange so let's look at it in more detail.
# 
# It is a dictionary with a tuple as the key and a function as the value. The key tuple `(n, k)` has the following
# meaning: `n` datasets are temporally matched together and then given in sets of `k` columns to the metric
# calculator. The metric calculator then gets a DataFrame with the columns ['ref', 'k1', 'k2' ...] and so on
# depending on the value of k. The value of `(2, 2)` makes sense here since we only have two datasets and all our
# metrics also take two inputs.
# 
# This can be used in more complex scenarios to e.g. have three input datasets that are all temporally matched
# together and then combinations of two input datasets are given to one metric calculator while all three datasets
# are given to another metric calculator. This could look like this:
# 
# ```python
# { (3 ,2): metric_calc,
#   (3, 3): triple_collocation}
# ```
# 
# Create the variable ***save_path*** which is a string representing the path where the results will be saved.
# **DO NOT CHANGE** the name ***save_path*** because it will be searched during the parallel processing!
# In[9]:

save_path = output_folder

import pprint
for job in jobs:
    
    results = process.calc(*job)
    pprint.pprint(results)
    netcdf_results_manager(results, save_path)


# The validation is then performed by looping over all the defined jobs and storing the results.
# You can see that the results are a dictionary where the key is a tuple defining the exact combination of datasets
# and columns that were used for the calculation of the metrics. The metrics itself are a dictionary of `metric-name:
#  numpy.ndarray` which also include information about the gpi, lon and lat. Since all the information contained in
# the job is given to the metric calculator they can be stored in the results.
# 
# Storing of the results to disk is at the moment supported by the `netcdf_results_manager` which creates a netCDF
# file for each dataset combination and stores each metric as a variable. We can inspect the stored netCDF file which
#  is named after the dictionary key:

# In[10]:


import netCDF4
results_fname = os.path.join(save_path, 'ASCAT.sm_with_ISMN.soil moisture.nc')

with netCDF4.Dataset(results_fname) as ds:
    for var in ds.variables:
        print(var, ds.variables[var][:])


# ## Parallel processing
# 
# The same code can be executed in parallel by defining the following `start_processing` function.

# In[11]:


def start_processing(job):
    try:
        return process.calc(*job)
    except RuntimeError:
        return process.calc(*job)


# `pytesmo.validation_framework.start_validation` can then be used to run your validation in parallel.
# Your setup code can look like this Ipython notebook without the loop over the jobs. Otherwise the validation would
# be done twice. Save it into a `.py` file e.g. `my_validation.py`.
# 
# After [starting the ipyparallel cluster](http://ipyparallel.readthedocs.org/en/latest/process.html) you can then
# execute the following code:
# 
# ```python
# from pytesmo.validation_framework import start_validation
# 
# # Note that before starting the validation you must start a controller
# # and engines, for example by using: ipcluster start -n 4
# # This command will launch a controller and 4 engines on the local machine.
# # Also, do not forget to change the setup_code path to your current setup.
# 
# setup_code = "my_validation.py"
# start_validation(setup_code)
# ```
# 
# ## Masking datasets
# 
# Masking datasets are datasets that return a pandas DataFrame with boolean values. `True` means that the observation
#  should be masked, `False` means it should be kept. All masking datasets are temporally matched in pairs to the
# temporal reference dataset. Only observations for which all masking datasets have a value of `False` are kept for
# further validation.
# 
# The masking datasets have the same format as the dataset dictionary and can be specified in the Validation class
# with the `masking_datasets` keyword.
# 
# ### Masking adapter
# 
# To easily transform an existing dataset into a masking dataset `pytesmo` offers a adapter class that calls the
# `read_ts` method of an existing dataset and creates a masking dataset based on an operator, a given threshold, and (optionally) a column name.

# In[12]:


from pytesmo.validation_framework.adapters import MaskingAdapter

ds_mask = MaskingAdapter(ismn_reader, '<', 0.2, 'soil moisture')
print(ds_mask.read_ts(ids[0]).head())


# ### Self-masking adapter
# `pytesmo` also has a class that masks a dataset "on-the-fly", based on one of the columns it contains and an operator and a threshold. In contrast to the masking adapter mentioned above, the output of the self-masking adapter is the masked data, not the the mask. The self-masking adapter wraps a data reader, which must have a `read_ts` or `read` method. Calling its `read_ts`/`read` method will return the masked data - more precisely a DataFrame with only rows where the masking condition is true.

# In[13]:


from pytesmo.validation_framework.adapters import SelfMaskingAdapter

ds_mask = SelfMaskingAdapter(ismn_reader, '<', 0.2, 'soil moisture')
print(ds_mask.read_ts(ids[0]).head())




"""
Module to load color maps from json files.
This module is deprecated and will be removed soon.
Please use the colorella package instead:
Website: https://github.com/TUW-GEO/colorella
To install colorella, run 'pip install colorella'
"""

import matplotlib.cm as cm
import matplotlib.colors as colors
import os
import json
import glob
from pytesmo.utils import deprecated


@deprecated()
def colormaps_path():
    """Returns application's default path for storing user-defined colormaps"""
    return os.path.dirname(__file__)


@deprecated()
def get_system_colormaps():
    """Returns the list of colormaps that ship with matplotlib"""
    return [m for m in cm.datad]


@deprecated()
def get_user_colormaps(cmap_fldr=colormaps_path()):
    """Returns a list of user-defined colormaps in the specified folder
    (defaults to standard colormaps folder if not specified)."""
    user_colormaps = []
    for root, dirs, files in os.walk(cmap_fldr):
        files = glob.glob(root + "/*.cmap")
        for name in files:
            with open(os.path.join(root, name), "r") as fidin:
                cmap_dict = json.load(fidin)
                user_colormaps.append(cmap_dict.get("name", name))
    return user_colormaps


@deprecated()
def load_colormap(json_file):
    """Generates and returns a matplotlib colormap from the specified JSON file,
    or None if the file was invalid."""
    colormap = None
    with open(json_file, "r") as fidin:
        cmap_dict = json.load(fidin)
        if cmap_dict.get("colors", None) is None:
            return colormap
        colormap_type = cmap_dict.get("type", "linear")
        colormap_name = cmap_dict.get("name", os.path.basename(json_file))
        if colormap_type == "linear":
            colormap = colors.LinearSegmentedColormap.from_list(
                name=colormap_name, colors=cmap_dict["colors"]
            )
        elif colormap_type == "list":
            colormap = colors.ListedColormap(
                name=colormap_name, colors=cmap_dict["colors"]
            )
    return colormap


@deprecated()
def load(cmap_name, cmap_folder=colormaps_path()):
    """Returns the matplotlib colormap of the specified name -
    if not found in the predefined
    colormaps, searches for the colormap in the specified
    folder (defaults to standard colormaps
    folder if not specified)."""
    cmap_name_user = cmap_name + ".cmap"
    user_colormaps = get_user_colormaps(cmap_folder)
    system_colormaps = get_system_colormaps()

    if cmap_name_user in user_colormaps:
        cmap_file = os.path.join(cmap_folder, cmap_name_user)
        cmap = load_colormap(cmap_file)
    elif cmap_name in system_colormaps:
        return cm.get_cmap(cmap_name)
    else:
        raise ValueError("Colormap not found")
    return cmap



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
Test datasets used in several tests
'''

import numpy as np
import numpy.testing as nptest
import pandas as pd

import pygeogrids.grids as grids
from pygeobase.io_base import GriddedTsBase


class TestDataset(object):
    """Test dataset that acts as a fake object for the base classes."""
    __test__ = False # prevent pytest from collecting this class during testing

    def __init__(self, filename, mode='r', n=1000):
        self.filename = filename
        self.mode = mode
        self.n = n

    def read(self, *args, **kwargs):

        if 'start' in kwargs:
            start = kwargs['start']
        else:
            start = '2000-01-01'

        x = np.arange(self.n) * 1.0  # must be a float
        y = np.arange(self.n) * 0.5
        index = pd.date_range(start=start, periods=self.n, freq="D")

        df = pd.DataFrame({'x': x, 'y': y}, columns=['x', 'y'], index=index)
        return df

    def alias_read(self, *args, **kwargs):
        return self.read(*args, **kwargs)

    def write(self, gpi, data):
        return None

    def read_ts(self, *args, **kwargs):
        return self.read(*args, **kwargs)

    def read_ts_other(self, *args, **kwargs):
        return self.read(*args, **kwargs) * 2

    def write_ts(self, gpi, data):
        return None

    def close(self):
        pass

    def flush(self):
        pass


class MaskingTestDataset(TestDataset):

    def read(self, *args, **kwargs):
        limit = kwargs.pop("limit")
        data = super(MaskingTestDataset, self).read(*args)
        data = data[['x']]
        data = data < limit
        data = data[:limit]
        return data


def test_masking_testdataset():

    ds = MaskingTestDataset("")
    data = ds.read(1, limit=500)
    data_should = np.ones((500), dtype=bool)
    nptest.assert_almost_equal(data['x'].values, data_should)
    data = ds.read(1, limit=250)
    data_should = np.ones((250), dtype=bool)
    nptest.assert_almost_equal(data['x'].values, data_should)


def setup_TestDatasets() -> dict:
    grid = grids.CellGrid(np.array([1, 2, 3, 4]), np.array([1, 2, 3, 4]),
                          np.array([4, 4, 2, 1]), gpis=np.array([1, 2, 3, 4]))

    ds1 = GriddedTsBase("", grid, TestDataset)
    ds2 = GriddedTsBase("", grid, TestDataset)
    ds3 = GriddedTsBase("", grid, TestDataset)

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['x'],
            'args': [],
            'kwargs': {}
        },
        'DS2': {
            'class': ds2,
            'columns': ['y'],
            'args': [],
            'kwargs': {},
            'use_lut': False,
            'grids_compatible': True
        },
        'DS3': {
            'class': ds3,
            'columns': ['x', 'y'],
            'args': [],
            'kwargs': {},
            'use_lut': False,
            'grids_compatible': True
        }
    }

    return datasets


def setup_two_without_overlap():
    grid = grids.CellGrid(np.array([1, 2, 3, 4]), np.array([1, 2, 3, 4]),
                          np.array([4, 4, 2, 1]), gpis=np.array([1, 2, 3, 4]))

    ds1 = GriddedTsBase("", grid, TestDataset)
    ds2 = GriddedTsBase("", grid, TestDataset)

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['x'],
            'args': [],
            'kwargs': {}
        },
        'DS2': {
            'class': ds2,
            'columns': ['y'],
            'args': [],
            'kwargs': {'start': '1990-01-01'},
            'use_lut': False,
            'grids_compatible': True
        }
    }
    return datasets


def setup_three_with_two_overlapping():
    grid = grids.CellGrid(np.array([1, 2, 3, 4]), np.array([1, 2, 3, 4]),
                          np.array([4, 4, 2, 1]), gpis=np.array([1, 2, 3, 4]))

    ds1 = GriddedTsBase("", grid, TestDataset)
    ds2 = GriddedTsBase("", grid, TestDataset)
    ds3 = GriddedTsBase("", grid, TestDataset)

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['x'],
            'args': [],
            'kwargs': {}
        },
        'DS2': {
            'class': ds2,
            'columns': ['y'],
            'args': [],
            'kwargs': {'start': '1990-01-01'},
            'use_lut': False,
            'grids_compatible': True
        },
        'DS3': {
            'class': ds3,
            'columns': ['x', 'y'],
            'args': [],
            'kwargs': {},
            'use_lut': False,
            'grids_compatible': True
        }
    }
    return datasets
